"""
This file contains the actions supported by pycompss-cli.
They are invoked from cli/pycompss.py and uses core/cmd.py.
"""
from abc import ABC, abstractmethod
import json
import os
import shutil
from pathlib import Path
from glob import glob
from pycompss_cli.core import utils
from uuid import uuid4 as uuid

class Actions(ABC):

    def __init__(self, arguments, debug=False, env_conf=None) -> None:
        super().__init__()
        self.arguments = arguments
        self.debug = debug
        self.env_conf = env_conf
        self.home_path = str(Path.home())
        
        if self.env_conf:
            self.env_conf['env_path'] = self.home_path + '/.COMPSs/envs/' + self.env_conf['name']

        if 'name' in self.arguments and self.arguments.name == 'unique uuid':
            self.arguments.name = ''.join(str(uuid()).split('-')[:2])

    
    def env_add_conf(self, extra_conf):
        current_env, env_conf_path = utils.get_current_env_conf(return_path=True)
        new_conf =  {**current_env, **extra_conf}
        with open(env_conf_path, 'w') as f:
            json.dump(new_conf, f)


    @abstractmethod
    def init(self):
        if not self.arguments.env:
            self.arguments.func()
            exit(0)

        del self.arguments.func

        if os.path.isdir(self.home_path + '/.COMPSs/envs'):
            envs = os.listdir(self.home_path + '/.COMPSs/envs')

            if self.arguments.name in envs:
                print("ERROR: There's already another environment named " + self.arguments.name)
                exit(1)

        env_path = self.home_path + '/.COMPSs/envs/' + self.arguments.name
        self.arguments.env_path = env_path

        os.makedirs(env_path)

        with open(env_path + '/env.json', 'w') as env_conf:
            json.dump(vars(self.arguments), env_conf)
        self.env_conf = vars(self.arguments)

        if self.arguments.config:
            shutil.copy2(self.arguments.config, env_path)

        for current_file in glob(self.home_path + '/.COMPSs/envs/*/current'):
            os.remove(current_file)

        open(env_path + '/current', 'a').close()

        print('Environment ID:', self.arguments.name)

    @abstractmethod
    def exec(self):
        pass

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def job(self):
        if not self.arguments.job:
            self.arguments.func()
            exit(0)

    def job_submit(self):
        if not self.arguments.rest_args:
            print(self.arguments.enqueue_args)
            exit(0)

    @abstractmethod
    def app(self):
        raise NotImplementedError("Wrong Environment! Try switching to a `cluster` environment")

    @abstractmethod
    def jupyter(self):
        pass

    @abstractmethod
    def monitor(self):
        pass

    @abstractmethod
    def gengraph(self):
        pass

    @abstractmethod
    def components(self):
        pass
    
    def environment(self):
        action_name = 'list'

        if self.arguments.environment:
            action_name = self.arguments.environment

        action_name = utils.get_object_method_by_name(self, 'env_' + action_name, include_in_name=True)
        getattr(self, action_name)()

    def env_list(self):
        envs_path = self.home_path + '/.COMPSs/envs'
        envs_files = os.listdir(envs_path)

        env_info = []
        for env_name in envs_files:
            env_type = json.load(open(envs_path + f'/{env_name}/env.json'))['env']
            env_current = '*' if 'current' in os.listdir(envs_path + f'/{env_name}') else ''
            env_info.append([env_name, env_type, env_current])
        
        col_names = ['ID', 'Type', 'Active']
        utils.table_print(col_names, env_info)


    def env_change(self, env_id=None):
        env_id = self.arguments.env_id if env_id is None else env_id
        if not os.path.isdir(self.home_path + '/.COMPSs/envs/' + env_id):
            print("ERROR: There's no environment named " + env_id)
            exit(1)

        current_files = glob(self.home_path + '/.COMPSs/envs/*/current')
        if current_files:
            os.remove(current_files[0])
        env_dir_name = self.home_path + '/.COMPSs/envs/' + env_id
        open(env_dir_name + '/current', 'a').close()
        print(f'Environment `{env_id}` is now active')

    @abstractmethod
    def env_remove(self, eid=None):
        self.env_change(env_id='default')
        env_id = self.arguments.env_id if eid is None else eid

        env_dir_name = self.home_path + '/.COMPSs/envs/' + env_id
        
        print(f'Deleting environment `{env_id}`...')
        shutil.rmtree(env_dir_name)


import os
import sys
import argparse
import subprocess

from pycompss_cli.core.docker.arguments import docker_init_parser
from pycompss_cli.core.local.arguments import local_init_parser
from pycompss_cli.core.remote.arguments import cluster_init_parser, cluster_parser_job, cluster_parser_app
from pycompss_cli.core import utils

FORMATTER_CLASS = argparse.RawTextHelpFormatter

def parse_sys_argv():
    """ Parses the sys.argv.

    :returns: All arguments as namespace.
    """
    parser = argparse.ArgumentParser(formatter_class=FORMATTER_CLASS)
    parser.add_argument("-d", "--debug",
                        help="Enable debug mode. Overrides log_level",
                        action="store_true")

    # Parent parser - includes all arguments which are common to all actions
    parent_parser = argparse.ArgumentParser(add_help=False,
                                            formatter_class=FORMATTER_CLASS)
    # Action sub-parser
    subparsers = parser.add_subparsers(dest="action")
    # INIT
    parser_init = subparsers.add_parser("init",
                                        aliases=["i"],
                                        help="Initialize COMPSs environment (default local).",
                                        parents=[parent_parser],
                                        formatter_class=FORMATTER_CLASS)
                                        
    parser_init.set_defaults(func=lambda: print(parser_init.format_help()))
    
    parser_init.add_argument("-cfg", "--config",
                             default="",
                             type=str,
                             help="Configuration file")

    parser_init.add_argument("-n", "--name",
                             default='unique uuid',
                             type=str,
                             help="Environment name")

    init_env_subparser = parser_init.add_subparsers(title="environment", dest="env")
    # init_env_subparser.default = "local"

    init_env_subparser.add_parser("docker", add_help=False, 
                                    parents=[docker_init_parser()])

    init_env_subparser.add_parser("local", add_help=False, 
                                    parents=[local_init_parser()])

    init_env_subparser.add_parser("cluster", add_help=False, 
                                    parents=[cluster_init_parser()])

    # EXEC
    parser_exec = subparsers.add_parser("exec",
                                        aliases=["ex"],
                                        help="Execute the given command within the COMPSs\' environment.",  # noqa: E501
                                        parents=[parent_parser],
                                        formatter_class=FORMATTER_CLASS)
    parser_exec.set_defaults(action='exec')

    
    # parser_exec.add_argument("-eid", "--env_id",
    #                          default="",
    #                          type=str,
    #                          help="Environment ID")
    parser_exec.add_argument('exec_cmd', 
                            nargs=argparse.REMAINDER,   
                            help="Exec program arguments")
    # RUN
    parser_run = subparsers.add_parser("run",
                                       aliases=["r"],
                                       help="Run the application (with runcompss) within the COMPSs\' environment.",  # noqa: E501
                                       parents=[parent_parser],
                                       formatter_class=FORMATTER_CLASS)
    parser_run.set_defaults(action='run')

    if utils.check_exit_code('runcompss -h') == 0:
        parser_run.epilog = subprocess.check_output('runcompss -h', shell=True).decode()
    else:
        assets_folder = os.path.dirname(os.path.abspath(__file__)) + '/..'
        with open(assets_folder + '/assets/runcompss_args.txt', 'r', encoding='utf-8') as f:
            parser_run.epilog = f.read()

    # parser_run.add_argument("-eid", "--env_id",
    #                          default="",
    #                          type=str,
    #                          help="Environment ID")

    parser_run.add_argument("-app", "--app_name",
                             default="",
                             type=str,
                             help="Name of the app where to execute runcompss. Only required for `cluster` type environment")
    parser_run.add_argument('rest_args',
                            nargs=argparse.REMAINDER,   
                            help="Runcompss program arguments")

    # APP  cluster_parser_app
    parser_app = subparsers.add_parser("app", aliases=["a"], add_help=False,
                                    help="Manage applications within cluster environments.",  # noqa: E501
                                    parents=[cluster_parser_app()],
                                    formatter_class=FORMATTER_CLASS)
    parser_app.set_defaults(action='app')


    # JOB
    parser_job = subparsers.add_parser("job", aliases=["j"], add_help=False,
                                    help="Manage jobs within cluster environments.",  # noqa: E501
                                    parents=[cluster_parser_job()],
                                    formatter_class=FORMATTER_CLASS)
    parser_job.set_defaults(action='job')
    

    # parser_job.add_argument("-eid", "--env_id",
    #                          default="",
    #                          type=str,
    #                          help="Environment ID")

    # MONITOR
    parser_monitor = subparsers.add_parser("monitor",
                                           aliases=["m"],
                                           help="Start the monitor within the COMPSs\' environment.",  # noqa: E501
                                           parents=[parent_parser],
                                           formatter_class=FORMATTER_CLASS)
    parser_monitor.set_defaults(action='monitor')
    
    parser_monitor.add_argument("option",
                                help="Start or stop de monitoring service.",
                                choices=["start", "stop"],
                                default="start",
                                type=str)

    # JUPYTER
    parser_jupyter = subparsers.add_parser("jupyter",
                                           aliases=["jpy"],
                                           help="Starts Jupyter within the COMPSs\' environment.",  # noqa: E501
                                           parents=[parent_parser],
                                           formatter_class=FORMATTER_CLASS)
    parser_jupyter.set_defaults(action='jupyter')
    
    # parser_jupyter.add_argument("-eid", "--env_id",
    #                          default="",
    #                          type=str,
    #                          help="Environment ID")
    parser_jupyter.add_argument("-app", "--app_name",
                             default="",
                             type=str,
                             help="Name of the app where the notebook will be deployed. Only required for `cluster` type environment")
    parser_jupyter.add_argument('rest_args', 
                            nargs=argparse.REMAINDER,   
                            help="Jupyter arguments")
    # GENGRAPH
    parser_gengraph = subparsers.add_parser("gengraph",
                                            aliases=["g"],
                                            help="Converts the given graph into pdf.",  # noqa: E501
                                            parents=[parent_parser],
                                            formatter_class=FORMATTER_CLASS)
    parser_gengraph.set_defaults(action='gengraph')

    # parser_gengraph.add_argument("-eid", "--env_id",
    #                          default="",
    #                          type=str,
    #                          help="Environment ID")                                            
    parser_gengraph.add_argument("dot_file",
                                 type=str,
                                 help="Dot file to convert to pdf")
    # COMPONENTS
    parser_components = subparsers.add_parser("components",
                                              aliases=["c"],
                                              help="Manage infrastructure components.",  # noqa: E501
                                              parents=[parent_parser],
                                              formatter_class=FORMATTER_CLASS)
    parser_components.set_defaults(action='components')

    # parser_components.add_argument("-eid", "--env_id",
    #                          default="",
    #                          type=str,
    #                          help="Environment ID")                                                  
    subparsers_components = parser_components.add_subparsers(dest="components")
    
    subparsers_components.add_parser("list",
                                    aliases=["l"],
                                    help="List COMPSs active components.",  # noqa: E501
                                    formatter_class=FORMATTER_CLASS)        # noqa: E501
    parser_components_add = subparsers_components.add_parser("add",
                                                             aliases=["a"],
                                                             help="Adds the RESOURCE to the pool of workers of the COMPSs.",  # noqa: E501
                                                             formatter_class=FORMATTER_CLASS)                                 # noqa: E501
    subparsers_components_add = parser_components_add.add_subparsers(dest="add")                          # noqa: E501
    parser_components_add_worker = subparsers_components_add.add_parser("worker",                         # noqa: E501
                                                                        aliases=["w"],                    # noqa: E501
                                                                        help="Add a worker.",             # noqa: E501
                                                                        formatter_class=FORMATTER_CLASS)  # noqa: E501
    parser_components_add_worker.add_argument("worker",
                                              type=str,
                                              default="1",
                                              help="Number of workers to add (can be integer or <IP>:<CORES> to add remote workers).")  # noqa: E501
    parser_components_remove = subparsers_components.add_parser("remove",
                                                                aliases=["r"],
                                                                help="Removes the RESOURCE to the pool of workers of the COMPSs.",  # noqa: E501
                                                                formatter_class=FORMATTER_CLASS)                                    # noqa: E501
    subparsers_components_remove = parser_components_remove.add_subparsers(dest="remove")                                           # noqa: E501
    parser_components_remove_worker = subparsers_components_remove.add_parser("worker",                                             # noqa: E501
                                                                              aliases=["w"],                                        # noqa: E501
                                                                              help="Remove a worker.",                              # noqa: E501
                                                                              formatter_class=FORMATTER_CLASS)                      # noqa: E501
    parser_components_remove_worker.add_argument("worker",
                                                 type=str,
                                                 default="1",
                                                 help="Number of workers to remove (can be integer or <IP>:<CORES> to add remote workers).")  # noqa: E501


    # ENVIRONMENT
    parser_environment = subparsers.add_parser("environment",
                                              aliases=["env"],
                                              help="Manage COMPSs environments.",  # noqa: E501
                                              parents=[parent_parser],
                                              formatter_class=FORMATTER_CLASS)

    parser_environment.set_defaults(action='environment')

    subparsers_environment = parser_environment.add_subparsers(dest="environment")

    subparsers_environment.add_parser("list",
                                        aliases=["l"],
                                        help="List COMPSs active environments.",  # noqa: E501
                                        formatter_class=FORMATTER_CLASS)        # noqa: E501
    parser_environment_change = subparsers_environment.add_parser("change",
                                                             aliases=["c"],
                                                             help="Change current COMPSs environment.",  # noqa: E501
                                                             formatter_class=FORMATTER_CLASS)
    parser_environment_change.add_argument("env_id",
                                            type=str,
                                            help="ID of the environment to set as active")  # noqa: E501                 

    parser_environment_remove = subparsers_environment.add_parser("remove",
                                                                aliases=["r"],
                                                                help="Removes COMPSs environment.",  # noqa: E501
                                                                formatter_class=FORMATTER_CLASS)                  # noqa: E501

    parser_environment_remove.add_argument("-f", "--force",
                             action='store_true',
                             default=False,
                             help="Force deleting de environment and the applications") 

    parser_environment_remove.add_argument("env_id",
                                            nargs="+",
                                            type=str,
                                            help="ID of the environment to remove")  # noqa: E501                                                                           

    # Check if the user does not include any argument
    if len(sys.argv) < 2:
        print(parser.print_help())
        sys.exit(1)

    arguments, leftovers = parser.parse_known_args()
    if leftovers:
        arguments.rest_args = leftovers + arguments.rest_args

    return arguments



import pycompss_cli.core.utils as utils
import subprocess
from typing import List

from pycompss_cli.core.cmd_helpers import command_runner

# ################ #
# GLOBAL VARIABLES #
# ################ #

# ############# #
# API FUNCTIONS #
# ############# #

def local_deploy_compss(working_dir: str = "") -> None:
    """ Starts the main COMPSs image in Docker.
    It stops any existing one since it can not coexist with itself.

    :param working_dir: Given working directory
    :param image: Given docker image
    :param restart: Force stop the existing and start a new one.
    :returns: None
    """
    
    # cfg_content = '{"working_dir":"' + working_dir + \
    #                 '","resources":"","project":""}'
    # tmp_path, cfg_file = _store_temp_cfg(cfg_content)
    # _copy_file(cfg_file, default_cfg)
    # shutil.rmtree(tmp_path)
    pass


def local_run_app(cmd: List[str]) -> None:
    """ Execute the given command in the main COMPSs image in Docker.

    :param cmd: Command to execute.
    :returns: The execution stdout.
    """

    if utils.check_exit_code('which enqueue_compss') == 1:
        cmd = ['module load COMPSs'] + cmd
    cmd = ';'.join(cmd)

    subprocess.run(cmd, shell=True)


def local_jupyter(work_dir, jupyter_args):
    cmd = 'jupyter notebook --notebook-dir=' + work_dir
    process = subprocess.Popen(cmd + ' ' + jupyter_args, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    try:
        while True:
            line = process.stdout.readline()
            if line == '' and process.poll() is not None:
                break
            if line:
                print(line.strip().decode('utf-8'))
            if process.poll() is not None:
                break
    except KeyboardInterrupt:
        print('Closing jupyter...')
        process.kill()
    

def local_exec_app(command, return_process=False):
    p = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if return_process:
        return p
    return p.stdout.decode().strip()


def local_submit_job(app_args, env_vars):
    cmd = f'enqueue_compss {app_args}'
    if utils.check_exit_code('which enqueue_compss') == 1:
        cmd = 'module load COMPSs;' + cmd
        
    if env_vars:
        cmd = ' ; '.join([*[f'export {var}' for var in env_vars], cmd])

    p = local_exec_app(cmd, return_process=True)
    job_id = p.stdout.decode().strip().split('\n')[-1].split(' ')[-1]
    if p.returncode != 0:
        print('ERROR:', p.stderr.decode())
    else:
        print('Job submitted:', job_id)
        return job_id

def local_job_list(local_job_scripts_dir):
    cmd = f"python3 {local_job_scripts_dir}/find.py"
    if utils.check_exit_code('which enqueue_compss') == 1:
        cmd = 'module load COMPSs;' + cmd
    return local_exec_app(cmd)

def local_cancel_job(local_job_scripts_dir, jobid):
    cmd = f"python3 {local_job_scripts_dir}/cancel.py {jobid}"
    if utils.check_exit_code('which enqueue_compss') == 1:
        cmd = 'module load COMPSs;' + cmd
    return local_exec_app(cmd)

def local_job_status(local_job_scripts_dir, jobid):
    cmd = f"python3 {local_job_scripts_dir}/status.py {jobid}"
    if utils.check_exit_code('which enqueue_compss') == 1:
        cmd = 'module load COMPSs;' + cmd
    status = local_exec_app(cmd)

    if status == 'SUCCESS\nSTATUS:':
        return 'ERROR'
    return status


import json
from glob import glob
from pathlib import Path
import subprocess
import time, os

def is_debug():
    return os.getenv('PYCOMPSS_CLI_DEBUG', 'false').lower() == 'true'

def get_object_method_by_name(obj, method_name, include_in_name=False):
    for class_method_name in dir(obj):
        if not '__' in class_method_name and callable(getattr(obj, class_method_name)):
            if class_method_name.startswith(method_name) or (include_in_name and method_name in class_method_name):
                return class_method_name

def table_print(col_names, data):
    col_widths = list(map(len, [item for sublist in data for item in sublist if item is not None]))
    if not col_widths:
        col_widths = [14]
    max_len = max(col_widths) + 1
    row_format ="{:>{l}}" * (len(col_names) + 1)
    print(row_format.format("", l=max_len, *col_names))
    for row in data:
        print(row_format.format('-', l=max_len, *row))

def get_current_env_conf(return_path=False):
    home_path = str(Path.home())
    current_env = glob(home_path + '/.COMPSs/envs/*/current')[0].replace('current', 'env.json')
    with open(current_env, 'r') as env:
        if return_path:
            return json.load(env), current_env
        return json.load(env)

def get_env_conf_by_name(env_name):
    home_path = str(Path.home())
    env_path = home_path + '/.COMPSs/envs/' + env_name + '/env.json'
    with open(env_path, 'r') as env:
        return json.load(env)

def ssh_run_commands(login_info, commands, **kwargs):
    cmd = ' ; '.join(commands)
    res = subprocess.run(f"ssh {login_info} '{cmd}'", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)
    return res.stdout.decode()

def check_exit_code(command):
    return subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).returncode

def poll_directory(directory, sleep_time=0.1):
    files = []
    while True:
        new_query = os.listdir(directory)
        if new_query != files:
            files = new_query
            yield set(files)
        time.sleep(sleep_time)


import json
import os
import sys
import tarfile
import tempfile
import shutil
from uuid import uuid4
import subprocess

from pycompss_cli.core.cmd_helpers import command_runner

# ################ #
# GLOBAL VARIABLES #
# ################ #

master_name = "pycompss-master"
worker_name = "pycompss-worker"
service_name = "pycompss-service"
default_workdir = "/home/user/"
default_worker_workdir = "/home/user/.COMPSsWorker"
default_cfg_file = "cfg"
default_cfg = default_workdir + "/" + default_cfg_file
default_image_file = "image"
default_image = default_workdir + "/" + default_image_file


IMAGE_NAME = "compss/compss:2.10"  # Update when releasing new version
DOCKER_AVAILABALE = True

try:
    import docker
    from docker.types import Mount
    from docker.errors import DockerException
except ImportError:
    DOCKER_AVAILABALE = False

# ############# #
# API FUNCTIONS #
# ############# #

class ErrorContainerNotRunning(Exception):
    pass

class DockerCmd(object):
    def __init__(self, env_id) -> None:
        global master_name
        global worker_name

        self.env_id = env_id

        if not DOCKER_AVAILABALE:
            print('ERROR: Pip package `docker` is required for creating docker environments.')
            exit(1)
        
        super().__init__()

        try:
            docker.APIClient(base_url="unix://var/run/docker.sock")
            self.client = docker.from_env()
        except DockerException:
            print("ERROR: Docker service is not running", file=sys.stderr)
            print("       Please, start docker service and try again", file=sys.stderr)
            exit(1)

        self.__setup_image_name()
        master_name = master_name + "-" + self.env_id
        worker_name = worker_name + "-" + self.env_id

    def __setup_image_name(self):
        global IMAGE_NAME
        if os.environ.get("DEFAULT_DISLIB_DOCKER_IMAGE") is not None:
            # This environment variable will be defined by the dislib script.
            # It can be overriden by the COMPSS_DOCKER_IMAGE or the -i flag
            # when running init.
            IMAGE_NAME = os.environ["DEFAULT_DISLIB_DOCKER_IMAGE"]
        elif os.environ.get("COMPSS_DOCKER_IMAGE") is not None:
            # If specified in an environment variable, take it
            IMAGE_NAME = os.environ["COMPSS_DOCKER_IMAGE"]
        elif len(self.client.containers.list(filters={"name": master_name})) > 0:
            # Condition equivalent to: is_running(master_name):
            # But since it is undefined yet, we do it explicitly.
            # If exists in the file (means that has been defined with init)
            master = self.client.containers.list(filters={"name": master_name})[0]
            # Command equivalent to: master = _get_master()
            # But since it is undefined yet, we do it explicitly.
            IMAGE_NAME = master.image.attrs["RepoTags"][0]

    def docker_deploy_compss(self, working_dir,
                            image: str = "",
                            restart: bool = True) -> None:
        """ Starts the main COMPSs image in Docker.
        It stops any existing one since it can not coexist with itself.

        :param working_dir: Given working directory
        :param image: Given docker image
        :param restart: Force stop the existing and start a new one.
        :returns: None
        """
        

        if image:
            docker_image = image
        else:
            docker_image = IMAGE_NAME

        masters = self.client.containers.list(filters={"name": master_name},
                                        all=True)

        assert len(masters) < 2  # never should we run 2 masters

        if restart or self._exists(master_name):
            self.docker_kill_compss(False)

        if not self.is_running(master_name):
            print("Starting %s container in dir %s" % (master_name, working_dir))
            print("If this is your first time running PyCOMPSs it may take a " +
                "while because it needs to download the docker image. " +
                "Please be patient.")
            subprocess.run(f'docker pull {docker_image}', shell=True)
            mounts = self._get_mounts(user_working_dir=working_dir)
            ports = {"8888/tcp": 8888,  # required for jupyter notebooks
                    "8080/tcp": 8080}  # required for monitor
            m = self.client.containers.run(image=docker_image, name=master_name,
                                    mounts=mounts, detach=True, ports=ports)
            self._generate_resources_cfg(ips=["localhost"])
            self._generate_project_cfg(ips=["localhost"])
            # don't pass configs because they need to be  overwritten when adding
            # new nodes
            cfg_content = '{"working_dir":"' + working_dir + \
                        '","resources":"","project":""}'
            tmp_path, cfg_file = self._store_temp_cfg(cfg_content)
            self._copy_file(cfg_file, default_cfg)
            shutil.rmtree(tmp_path)

    
    def docker_start_compss(self):
        self.client.containers.get(master_name).start()


    def docker_update_image(self) -> None:
        """ Updates the default docker image.

        :returns: None
        """
        docker_image = "compss/compss:latest"
        print("Updating docker image: %s" % docker_image)
        if "COMPSS_DOCKER_IMAGE" in os.environ:
            docker_image = os.environ["COMPSS_DOCKER_IMAGE"]
            print("Found COMPSS_DOCKER_IMAGE environment variable: %s. Updating." %
                docker_image)
        else:
            print("COMPSS_DOCKER_IMAGE is unset or empty. Updating default docker image: %s" %  # noqa: E501
                docker_image)
        command_runner(["docker", "pull", docker_image])


    def docker_kill_compss(self, clean: bool = True) -> None:
        """ Stops all COMPSs images in Docker.

        :param clean: Force clean the generated files.
        :returns: None
        """
        if clean:
            # Clean the cfg file
            try:
                master = self._get_master()
                self._remove_cfg(master)
            except ErrorContainerNotRunning:
                print("WARNING: No master container running.")

        self._stop_by_name(master_name)
        self._stop_by_name(worker_name)


    def docker_exec_in_daemon(self, cmd: str, return_output=False) -> None:
        """ Execute the given command in the main COMPSs image in Docker.

        :param cmd: Command to execute.
        :returns: The execution stdout.
        """

        if not self.is_running(master_name):
            self.docker_start_compss()

        master = self._get_master()
        _, output = master.exec_run(cmd, workdir=default_workdir, stream=True)
        if return_output:
            return list(output)[-1].decode().strip()
        try:
            for line in output:
                print(line.strip().decode())
        except KeyboardInterrupt:
            master.exec_run('compss_clean_procs')


    def docker_start_monitoring(self) -> None:
        """ Starts the COMPSs monitoring within the Docker instance.

        :returns: The monitoring initialization stdout.
        """
        print("Starting Monitor")
        if not self.is_running(master_name):
            self.start_daemon()

        cmd = "/etc/init.d/compss-monitor start"
        master = self._get_master()
        env = {"COMPSS_MONITOR": str(default_workdir) + "/.COMPSs"}
        _, output = master.exec_run(cmd,
                                    environment=env,
                                    workdir=default_workdir,
                                    stream=True)
        for line in output:
            print(line.strip().decode())
        print("Please, open: http://127.0.0.1:8080/compss-monitor")


    def docker_stop_monitoring(self) -> None:
        """ Stops the COMPSs monitoring within the Docker instance.

        :returns: The monitoring stop stdout.
        """
        print("Stopping Monitor")
        cmd = "/etc/init.d/compss-monitor stop"
        master = self._get_master()
        _, output = master.exec_run(cmd,
                                    workdir=default_workdir,
                                    stream=True)
        for line in output:
            print(line.strip().decode())


    def docker_components(self, option: str = "list",
                        resource: str = "worker",
                        value: str = "1") -> None:
        """ Performs actions over the COMPSS docker instances deployed.

        :param option: Option to perform (supported: list, add and remove)
        :param element: Element to add or remove (not needed for list)
        :param value: Amount of elements to add or remove (not needed for list)
        :returns: None
        """
        if option == "list":
            masters = self.client.containers.list(filters={"name": master_name})
            workers = self.client.containers.list(filters={"name": worker_name})
            for c in masters + workers:
                print(c.name)
        elif option == "add":
            if resource == "worker":
                if value.isdigit():
                    self._add_workers(int(value))
                else:
                    self._add_custom_worker(value)
            else:
                raise Exception("Unsupported resource to be added: " + resource)
        elif option == "remove":
            if resource == "worker":
                if value.isdigit():
                    self._remove_workers(int(value))
                else:
                    self._remove_custom_worker(value)
            else:
                raise Exception("Unsupported resource to be removed: " + resource)
        else:
            raise Exception("Unexpected components option: " + option)


    def is_running(self, name: str = master_name) -> bool:
        """ Checks if a docker instance is running.

        :param name: Instance name.
        :returns: True if running. False otherwise.
        """
        cs = self.client.containers.list(filters={"name": name})
        return len(cs) > 0

    def exists(self, name: str = master_name) -> bool:
        """ Checks if a docker instance exists.

        :param name: Instance name.
        :returns: True if exists. False otherwise.
        """
        cs = self.client.containers.list(all=True, filters={"name": name})
        return len(cs) > 0


    # ################# #
    # PRIVATE FUNCTIONS #
    # ################# #



    def _exists(self, name: str) -> bool:
        """ Checks if a docker instance exists.

        :param name: Instance name.
        :returns: True if exists. False otherwise.
        """
        cs = self.client.containers.list(filters={"name": name}, all=True)
        return len(cs) > 0


    def _get_master(self):
        """ Retrieve the COMPSs master container object.

        :returns: Master container object.
        """
        try:
            master = self.client.containers.list(filters={"name": master_name})[0]
        except IndexError:
            raise ErrorContainerNotRunning(master_name)
        return master


    def _get_workers(self) -> list:
        """ Retrieve the COMPSs worker containers objects.

        :returns: List of the Worker containers objects.
        """
        try:
            workers = self.client.containers.list(filters={"name": worker_name})
        except IndexError:
            raise ErrorContainerNotRunning(worker_name)
        return workers


    def _get_worker_ips(self) -> list:
        """ Retrieve the COMPSs worker containers IP address.

        :returns: List of the Worker containers IP address.
        """
        ips = [c.attrs["NetworkSettings"]["Networks"]["bridge"]["IPAddress"]
            for c in self.client.containers.list(filters={"name": worker_name})]
        return ips


    def _store_temp_cfg(self, cfg_content: str) -> tuple:
        """ Stores the given content in the temporary cfg file.

        :param cfg_content: Cfg file contents.
        :returns: The tmp file path and the cfg file name.
        """
        tmp_path = tempfile.mkdtemp()
        cfg_file = os.path.join(tmp_path, default_cfg_file)
        with open(cfg_file, "w") as f:
            f.write(cfg_content)
        return tmp_path, cfg_file


    def _copy_file(self, src: str, dst: str) -> None:
        """ Copy the given file to the given destination within the COMPSs docker
        master instance.

        :param src: Source file path.
        :param dst: Destination file path within the docker instance.
        :returns: None
        """
        master = self._get_master()
        os.chdir(os.path.dirname(src))
        src_name = os.path.basename(src)
        tar_name = src + ".tar"
        tar = tarfile.open(tar_name, mode="w")
        try:
            tar.add(src_name)
        finally:
            tar.close()
        data = open(tar_name, "rb").read()
        output = master.put_archive(os.path.dirname(dst), data)
        if not output:
            print("ERROR COPYING " + str(src) +
                " TO " + src(dst) +
                " OF MASTER CONTAINER!!!")


    def _get_mounts(self, user_working_dir: str) -> list:
        """ Retrieve the list of folders to be mounted. It gets the Mount object
        from the given user working directory, and can include any other needed
        folder.

        :param user_working_dir: User working path.
        :returns: List of docker Mount objects.
        """
        # mount target dir needs to be absolute
        target_dir = default_workdir
        user_dir = Mount(target=target_dir,
                        source=user_working_dir,
                        type="bind")
        # WARNING: mounting .COMPSs makes it fail
        # compss_logs_dir = os.environ["HOME"] + f"/.COMPSs"
        # os.makedirs(compss_logs_dir, exist_ok=True)

        # compss_log_dir = Mount(target="/root/.COMPSs",
        #                        source=compss_logs_dir,
        #                        type="bind")
        mounts = [user_dir] #, compss_log_dir]
        return mounts


    def _generate_project_cfg(self, ips: list = (), cpus: int = 4,
                            install_dir: str = "/opt/COMPSs",
                            worker_dir: str = default_worker_workdir) -> str:
        """ Generates the project.xml according to the given parameters.

        :param ips: List of ip of the worker nodes.
        :param cpus: Number of cores per worker node.
        :param install_dir: COMPSs installation directory.
        :param worker_dir: Worker working directory.
        :returns: The cfg file contents for the project.
        """
        # ./generate_project.sh project.xml "172.17.0.3:4:/opt/COMPSs:/tmp"
        master = self._get_master()
        proj_cmd = "/opt/COMPSs/Runtime/scripts/system/xmls/generate_project.sh"
        master_ip = "127.0.0.1"
        workers_ip = ips
        proj_master = ":".join((master_ip, "0", install_dir, worker_dir))
        proj_workers = " ".join(
            ["%s:%s:%s:%s" % (ip, cpus, install_dir, worker_dir) for ip in
            workers_ip])
        cmd = "%s /project.xml '%s' '%s'" % (proj_cmd, proj_master, proj_workers)
        exit_code, output = master.exec_run(cmd=cmd)
        if exit_code != 0:
            print("Exit code: %s" % exit_code)
            for line in [l for l in output.decode().split("\n")]:
                print(line)
            sys.exit(exit_code)
        return proj_workers


    def _generate_resources_cfg(self, ips: list = (), cpus: int = 4) -> str:
        """ Generates the resources.xml according to the given parameters.

        :param ips: List of ip of the worker nodes.
        :param cpus: Number of cores per worker node.
        :returns: The cfg file contents for the resources.
        """
        # ./generate_resources.sh resources.xml "172.17.0.3:4"
        master = self._get_master()
        res_cmd = "/opt/COMPSs/Runtime/scripts/system/xmls/generate_resources.sh"
        res_arg = " ".join(["%s:%s" % (ip, cpus) for ip in ips])
        cmd = "%s /resources.xml '%s'" % (res_cmd, res_arg)
        exit_code, output = master.exec_run(cmd=cmd)
        if exit_code != 0:
            print("Exit code: %s" % exit_code)
            for line in [l for l in output.decode().split("\n")]:
                print(line)
            sys.exit(exit_code)
        return res_arg


    def _get_cfg(self, master) -> dict:
        """ Retrieve the cfg file contents as dictionary.

        :param master: Master docker instance object.
        :returns: CFG file contents as dictionary.
        """
        exit_code, output = master.exec_run(cmd="cat " + default_cfg)
        json_str = output.decode()
        cfg = json.loads(json_str)
        return cfg


    def _remove_cfg(self, master) -> dict:
        """ Remove the cfg file.

        :param master: Master docker instance object.
        :returns: None
        """
        exit_code, output = master.exec_run(cmd="rm " + default_cfg)
        if exit_code != 0:
            for line in output:
                print(line.strip().decode())


    def _update_cfg(self, master, cfg: dict, ips, cpus) -> None:
        """ Update the cfg file with the given parameters.

        :param master: Master docker instance object.
        :param ips: List of IP addresses of the worker nodes.
        :param cpus: Number of cores per node.
        :returns: None
        """
        # Generate project.xml
        new_proj_cfg = self._generate_project_cfg(ips=ips, cpus=cpus)
        # Generate resources.xml
        new_res_cfg = self._generate_resources_cfg(ips=ips, cpus=cpus)
        # Update the cfg_content
        cfg_content = '{"working_dir":"' + cfg['working_dir'] + \
                    '","resources":"' + new_res_cfg + \
                    '","project":"' + new_proj_cfg + '"}'
        tmp_path, cfg_file = self._store_temp_cfg(cfg_content)
        self._copy_file(cfg_file, default_cfg)
        shutil.rmtree(tmp_path)


    def _add_custom_worker(self, custom_cfg: str) -> None:
        """ Add custom worker to the cfg file.
        * custom_cfg = "ip:cpus"

        :param custom_cfg: Existing custom cfg file.
        :returns: None
        """
        ip, cpus = custom_cfg.split(":")
        master = self._get_master()
        cfg = self._get_cfg(master)
        # try to copy the master working dir to custom worker
        os.system("scp -r %s %s:/tmp" % (cfg["working_dir"], ip))
        ips = self._get_worker_ips()
        ips.append(ip)
        self._update_cfg(master, cfg, ips, cpus)
        print("Connected worker %s\n\tCPUs: %s" % (ip, cpus))


    def _remove_custom_worker(self, custom_cfg: str) -> None:
        """ Remove custom worker from the cfg file.
        * custom_cfg = "ip:cpus"

        :param custom_cfg: Existing custom cfg file.
        :returns: None
        """
        ip, cpus = custom_cfg.split(":")
        master = self._get_master()
        cfg = self._get_cfg(master)
        # Find the worker with the given ip
        workers = self._get_workers()
        for w in workers:
            w_ip = w.attrs["NetworkSettings"]["Networks"]["bridge"]["IPAddress"]
            if w_ip == ip:
                w.remove(force=True)
        ips = self._get_worker_ips()
        self._update_cfg(master, cfg, ips, cpus)
        print("Removed worker %s" % (ip))


    def _add_workers(self, num_workers: int = 1,
                    user_working_dir: str = "",
                    cpus: int = 4) -> None:
        """ Add COMPSs workers to the cfg file.

        :param num_workers: Number of workers to e added.
        :param user_working_dir: User working directory.
        :returns: None
        """
        master = self._get_master()
        cfg = self._get_cfg(master)
        mounts = self._get_mounts(user_working_dir=cfg["working_dir"])
        for _ in range(num_workers):
            worker_id = worker_name + "-" + uuid4().hex[:8]
            self.client.containers.run(image=IMAGE_NAME, name=worker_id,
                                mounts=mounts, detach=True, auto_remove=True)
        ips = self._get_worker_ips()
        self._update_cfg(master, cfg, ips, cpus)
        print("Started %s worker/s\n\tWorking dir: %s\n\tCPUs: %s" %
            (num_workers, user_working_dir, cpus))


    def _remove_workers(self, num_workers: int = 1,
                        cpus: int = 4) -> None:
        """ Removes COMPSs workers from the cfg file.

        :param num_workers: Number of workers to e added.
        :param cpus: Number of cores of the workers.
        :returns: None
        """
        master = self._get_master()
        cfg = self._get_cfg(master)
        workers = self._get_workers()
        to_remove = workers[:num_workers]
        for worker in to_remove:
            worker.remove(force=True)
        ips = self._get_worker_ips()
        self._update_cfg(master, cfg, ips, cpus)
        print("Removed " + str(num_workers) + " workers.")


    def _stop_by_name(self, name: str) -> None:
        """ Stop a docker instance by name.

        :param name: Name of the instance to be removed.
        :returns: None
        """
        containers = self.client.containers.list(filters={"name": name}, all=True)
        for c in containers:
            c.remove(force=True)


import argparse
import os
import subprocess

from pycompss_cli.core import utils

FORMATTER_CLASS = argparse.RawDescriptionHelpFormatter

def cluster_init_parser():
    """ Parses the sys.argv.

    :returns: All arguments as namespace.
    """
    parser = argparse.ArgumentParser(formatter_class=FORMATTER_CLASS)

    # Parent parser - includes all arguments which are common to all actions
    parent_parser = argparse.ArgumentParser(add_help=False,
                                            formatter_class=FORMATTER_CLASS)
    # Action sub-parser
    subparsers = parser.add_subparsers(dest="action")
    # INIT
    parser_init = subparsers.add_parser("init",
                                        aliases=["i"],
                                        help="Initialize COMPSs within a given cluster node.",
                                        parents=[parent_parser],
                                        formatter_class=FORMATTER_CLASS)

    parser_init.add_argument("-l", "--login",
                             type=str,
                             required=True,
                             help="Login info username@cluster_hostname")

    parser_init.add_argument("-m", "--modules",
                             nargs='*',
                             help="Module list or file to load in cluster")

    return parser_init

def cluster_parser_app():
    parser = argparse.ArgumentParser(formatter_class=FORMATTER_CLASS)

    # Parent parser - includes all arguments which are common to all actions
    parent_parser = argparse.ArgumentParser(add_help=False,
                                            formatter_class=FORMATTER_CLASS)
    # Action sub-parser
    subparsers = parser.add_subparsers(dest="action")

    # APP
    parser_app = subparsers.add_parser("app",
                                        aliases=["a"],
                                        parents=[parent_parser],
                                        formatter_class=FORMATTER_CLASS)
    parser_app.set_defaults(func=lambda: print(parser_app.format_help()))

    app_subparsers = parser_app.add_subparsers(dest="app")

    app_deploy_parser = app_subparsers.add_parser("deploy",
                                aliases=["d"],
                                help="Deploy an application to a cluster or remote environment",
                                parents=[parent_parser],
                                formatter_class=FORMATTER_CLASS)

    app_deploy_parser.add_argument("app_name",
                             type=str,
                             help="Name of the application")

    app_deploy_parser.add_argument("-ls", "--local_source",
                             default='current directory',
                             type=str,
                             help="Path from which the files will be copied. Can be a directory or a single file")

    app_deploy_parser.add_argument("-rd", "--remote_dir",
                             type=str,
                             help="Remote destination directory to copy the local app files")

    app_remove_parser = app_subparsers.add_parser("remove",
                                aliases=["r"],
                                help="Delete one or more deployed applications",
                                parents=[parent_parser],
                                formatter_class=FORMATTER_CLASS)

    app_remove_parser.add_argument("app_name",
                             type=str,
                             nargs='+',
                             help="Name of the application")

    app_subparsers.add_parser("list",
                                aliases=["l"],
                                help="List all deployed applications",
                                parents=[parent_parser],
                                formatter_class=FORMATTER_CLASS)

    return parser_app

def cluster_parser_job():
    """ Parses the sys.argv.

    :returns: All arguments as namespace.
    """
    parser = argparse.ArgumentParser(formatter_class=FORMATTER_CLASS)

    # Parent parser - includes all arguments which are common to all actions
    parent_parser = argparse.ArgumentParser(add_help=False,
                                            formatter_class=FORMATTER_CLASS)
    # Action sub-parser
    subparsers = parser.add_subparsers(dest="action")

    # JOB
    parser_job = subparsers.add_parser("job",
                                        aliases=["j"],
                                        parents=[parent_parser],
                                        formatter_class=FORMATTER_CLASS)
    parser_job.set_defaults(func=lambda: print(parser_job.format_help()))

    job_subparsers = parser_job.add_subparsers(dest="job")

    submit_job_parser = job_subparsers.add_parser("submit",
                                aliases=["sub"],
                                help="Submit a job to a cluster or remote environment",
                                parents=[parent_parser],
                                formatter_class=FORMATTER_CLASS)
    
    if utils.check_exit_code('enqueue_compss -h') == 0:
        enqueue_args = subprocess.check_output('enqueue_compss -h', shell=True).decode()
    else:
        assets_folder = os.path.dirname(os.path.abspath(__file__)) + '/../..'
        with open(assets_folder + '/assets/enqueue_compss_args.txt', 'r', encoding='utf-8') as f:
            enqueue_args = f.read()

    submit_job_parser.set_defaults(enqueue_args=enqueue_args)
    submit_job_parser.epilog = enqueue_args

    submit_job_parser.add_argument("--verbose", "-v",
                                    action='store_true',
                                    help="Shows the final command that is executed on the remote")

    submit_job_parser.add_argument("--env_var", "-e",
                                    action='append',
                                    nargs=1,
                                    type=str,
                                    default=[],
                                    help="Set environment variables")  # noqa: E501   

    submit_job_parser.add_argument("-app", "--app_name",
                             default="",
                             type=str,
                             help="Name of the app where to execute runcompss. Only required for `cluster` type environment")

    submit_job_parser.add_argument('rest_args', 
                            nargs=argparse.REMAINDER,   
                            help="Remote enqueue_compss arguments")

    status_parser = job_subparsers.add_parser("status",
                                aliases=["st"],
                                help="Check status of submitted job",
                                parents=[parent_parser],
                                formatter_class=FORMATTER_CLASS)

    status_parser.add_argument("job_id",
                             type=str,
                             help="Job ID to check status")

    cancel_parser = job_subparsers.add_parser("cancel",
                                aliases=["c"],
                                help="Cancel a submitted job",
                                parents=[parent_parser],
                                formatter_class=FORMATTER_CLASS)

    cancel_parser.add_argument("job_id",
                             type=str,
                             help="Job ID to cancel")

    job_subparsers.add_parser("list",
                                aliases=["l"],
                                help="List all the submitted jobs",
                                parents=[parent_parser],
                                formatter_class=FORMATTER_CLASS)

    history_parser = job_subparsers.add_parser("history",
                                aliases=["h"],
                                help="List all past submitted jobs and their app arguments",
                                parents=[parent_parser],
                                formatter_class=FORMATTER_CLASS)

    history_parser.add_argument("--job_id", '-j',
                             type=str,
                             help="")

    return parser_job


#!/usr/bin/python
#
#  Copyright 2002-2021 Barcelona Supercomputing Center (www.bsc.es)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
This script is aimed at checking the status of the running PyCOMPSs jobs
with Jupyter notebook.
"""

import sys
from commons import VERBOSE
from commons import SUCCESS_KEYWORD
from commons import ERROR_KEYWORD
from commons import get_job_status
from commons import setup_supercomputer_configuration


def status():
    """
    Main status function.
    Gets the status of the requested job identifier and prints it
    through stdout.
    :return: None
    """
    if VERBOSE:
        print("Checking status PyCOMPSs interactive job...")

    # Get command line arguments
    job_id = sys.argv[1]

    if VERBOSE:
        print(" - Job: " + str(job_id))

    # Load the Supercomputer configuration to get the appropriate status command
    setup_supercomputer_configuration()

    # Get the job status
    job_status, return_code = get_job_status(job_id)

    if VERBOSE:
        print("Finished checking status.")

    # Print to notify the status result
    if return_code != 0:
        print(ERROR_KEYWORD)
        exit(1)
    else:
        print(SUCCESS_KEYWORD)
        print("STATUS:" + job_status)


if __name__ == '__main__':
    status()





"""
This file contains the actions supported by pycompss-cli.
They are invoked from cli/pycompss.py and uses core/cmd.py.
"""

from collections import defaultdict
import datetime
import json
import traceback
from pycompss_cli.core.local.cmd import local_deploy_compss
from pycompss_cli.core.local.cmd import local_run_app
from pycompss_cli.core.local.cmd import local_exec_app
from pycompss_cli.core.local.cmd import local_jupyter
from pycompss_cli.core.local.cmd import local_submit_job
from pycompss_cli.core.local.cmd import local_job_list
from pycompss_cli.core.local.cmd import local_cancel_job
from pycompss_cli.core.local.cmd import local_job_status
from pycompss_cli.core.actions import Actions
import pycompss_cli.core.utils as utils
import os, sys


class LocalActions(Actions):
    def __init__(self, arguments, debug=False, env_conf=None) -> None:
        super().__init__(arguments, debug=debug, env_conf=env_conf)
        path = os.path.abspath(__file__)
        self.local_job_scripts_dir = os.path.dirname(path) + '/../remote/job_scripts'

        self.past_jobs = defaultdict(list)
        if self.env_conf and os.path.isfile(self.env_conf['env_path'] + '/jobs.json'):
            with open(self.env_conf['env_path'] + '/jobs.json') as f:
                self.past_jobs = defaultdict(list, json.load(f))

    def init(self):
        super().init()
        """ Deploys COMPSs infrastructure local env

        :param arguments: Command line arguments
        :param debug: Debug mode
        :returns: None
        """

   
        if self.arguments.working_dir == 'current directory':
            self.arguments.working_dir = os.getcwd()

        try:
            local_deploy_compss(self.arguments.working_dir)
        except:
            traceback.print_exc()
            print("ERROR: Local deployment failed")
            self.env_remove(self.arguments.name)


    def run(self):
        super().run()
        """ Run the given command in the COMPSs infrastructure at docker

        :param arguments: Command line arguments
        :param debug: Debug mode
        :returns: None
        """
        if self.debug:
            print("Running...")
            print("Parameters:")
            print("\t- Application: " + self.arguments.application)
            print("\t- Arguments: " + str(self.arguments.argument))

        app_args = self.arguments.rest_args

        commands = [
            "runcompss " + ' '.join(app_args)
        ]

        if 'working_dir' in self.env_conf:
            commands.insert(0, 'cd ' + self.env_conf['working_dir'])
                
        local_run_app(commands)

    def jupyter(self):
        if utils.check_exit_code('jupyter') == 127:
            print('ERROR: Jupyter not found!')
            exit(1)

        working_dir = '.'
        if 'working_dir' in self.env_conf:
            working_dir = self.env_conf['working_dir']

        jupyter_args = self.arguments.rest_args
        local_jupyter(working_dir, ' '.join(jupyter_args))

    def exec(self):
        command = ' '.join(self.arguments.exec_cmd)
        local_exec_app(command)

    def app(self):
        print("ERROR: Wrong Environment! Try switching to a `cluster` environment")
        exit(1)

    def env_remove(self, env_id=None):
        super().env_remove(eid=env_id)

    def job(self):
        super().job()

        action_name = self.arguments.job
        action_name = utils.get_object_method_by_name(self, 'job_' + action_name, include_in_name=True)
        getattr(self, action_name)()
                
    def job_submit(self):
        super().job_submit()

        app_name = None

        working_dir = os.getcwd()
        if 'working_dir' in self.env_conf:
            working_dir = self.env_conf['working_dir']

        for arg in reversed(self.arguments.rest_args):
            file_path = working_dir + '/' + arg
            if os.path.isfile(file_path):
                app_name = arg
                break

        if app_name is None:
            print('WARNING: No application found in the working directory!')

        app_args = ' '.join(self.arguments.rest_args)
        env_vars = [item for sublist in self.arguments.env_var for item in sublist]
        app_args = f'--pythonpath={working_dir} ' + app_args
        app_args = f'--classpath={working_dir} ' + app_args
        app_args = f'--appdir={working_dir} ' + app_args

        if self.arguments.verbose:
            print('\t-\tenvars:', env_vars)
            print('\t-\tenqueue_compss:', app_args)

        job_id = local_submit_job(app_args, env_vars)

        self.past_jobs[job_id] = {
                'app_name': app_name if app_name else 'error',
                'env_vars': '; '.join(env_vars) if env_vars else 'Empty',
                'enqueue_args': app_args,
                'timestamp': str(datetime.datetime.utcnow())[:-7] + ' UTC',
                'path': working_dir
            }

        with open(self.env_conf['env_path'] + '/jobs.json', 'w') as f:
            json.dump(self.past_jobs, f)

    def job_history(self):
        job_id = self.arguments.job_id
        if job_id:
            app_jobs = self.past_jobs[job_id]
            print('\tApp name:', app_jobs['app_name'])
            print('\tApp path:', app_jobs['path'])
            print('\tSubmit time:', app_jobs['timestamp'])
            print('\tEnvironment Variables:', app_jobs['env_vars'])
            print('\tEnqueue Args:', app_jobs['enqueue_args'])
            print()
        else:
            col_names = ['JobIDs', 'AppName', 'Status']
            rows = []
            for job_id, app_job in self.past_jobs.items():
                if 'status' not in app_job:
                    job_status = self.job_status(job_id)
                else:
                    job_status = app_job['status']
                    if 'COMPLETED' not in job_status:
                        job_status = self.job_status(job_id)
                if job_status is None:
                    job_status = 'Unknown'
                rows.append([job_id, app_job['app_name'], f'\t{job_status}'])
            utils.table_print(col_names, rows)

    def job_list(self):
        list_jobs = local_job_list(self.local_job_scripts_dir)
        print(list_jobs)
        return list_jobs

    def job_status(self, jid=None):
        job_id = self.arguments.job_id if jid is None else jid
        job_status = local_job_status(self.local_job_scripts_dir, job_id)
        if job_status == 'ERROR' and job_id in self.past_jobs:
            app_path = self.past_jobs[job_id]['path']
            cmd = f'grep -iF "error" {app_path}/compss-{job_id}.err'
            status = 'ERROR' if local_exec_app(cmd) else 'SUCCESS'
            job_status = 'COMPLETED:' + status

        if jid is None:
            print(job_status)

        self.past_jobs[job_id]['status'] = job_status
        with open(self.env_conf['env_path'] + '/jobs.json', 'w') as f:
            json.dump(self.past_jobs, f)

        return job_status

    def job_cancel(self):
        jobid = self.arguments.job_id
        res = local_cancel_job(self.local_job_scripts_dir, jobid)
        print(res)

    def monitor(self):
        if self.arguments.option == 'start':
            local_exec_app('/etc/init.d/compss-monitor start')
            local_exec_app('firefox http://localhost:8080/compss-monitor &')
        elif self.arguments.option == 'stop':
            local_exec_app('/etc/init.d/compss-monitor stop')
            
    def components(self):
        print('ERROR: Not Implemented Yet')
        exit(1)

    def gengraph(self):
        command = "compss_gengraph " + self.arguments.dot_file
        local_exec_app(command)



#!/usr/bin/python
#
#  Copyright 2002-2021 Barcelona Supercomputing Center (www.bsc.es)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
This script is aimed at getting the list of running PyCOMPSs jobs
with Jupyter notebook.
"""

import os
from commons import VERBOSE
from commons import SUCCESS_KEYWORD
from commons import ERROR_KEYWORD
from commons import command_runner
from commons import setup_supercomputer_configuration
from commons import get_job_name
from commons import get_job_status
from commons import verify_job_name


def find():
    """
    Main find function.
    Gets the jobs list command from the environment variable (defined in supercomputer cfgs)
    and runs it. It also gets each job identifier name and status to provide more information
    to the users.
    :return: None
    """
    if VERBOSE:
        print("Looking for PyCOMPSs interactive jobs...")

    # Load the Supercomputer configuration to get the appropriate status command
    setup_supercomputer_configuration()

    # Get the list of running job ids
    job_list_command = os.environ['QUEUE_JOB_LIST_CMD']
    return_code, raw_job_ids, _ = command_runner(job_list_command.split())
    job_ids = raw_job_ids.splitlines()

    # Filter the jobs (keep only the notebook related ones)
    jobs = []
    for job_id in job_ids:
        name = get_job_name(job_id).strip()
        status, _ = get_job_status(job_id)
        jobs.append((job_id, status, name))

    if VERBOSE:
        print("Finished looking for jobs.")

    # Print to notify the find result
    if return_code != 0:
        print(ERROR_KEYWORD)
        exit(1)
    else:
        # Print to provide the list of jobs to the client
        print(SUCCESS_KEYWORD)
        for job_id, status, name in jobs:
            print(str(job_id) + '\t- ' + str(status) + '\t- ' + str(name))


if __name__ == '__main__':
    find()



"""
This file contains the actions supported by pycompss-cli.
They are invoked from cli/pycompss.py and uses core/cmd.py.
"""

from pycompss_cli.core.actions import Actions
from pycompss_cli.core.docker.cmd import DockerCmd
from glob import glob
import os, traceback
 
class DockerActions(Actions):

    def __init__(self, arguments, debug=False, env_conf=None) -> None:
        super().__init__(arguments, debug=debug, env_conf=env_conf)
        env_id = arguments.name if env_conf is None else env_conf['name']
        self.docker_cmd = DockerCmd(env_id)


    def init(self):
        super().init()
        """ Deploys COMPSs infrastructure in Docker

        :param arguments: Command line arguments
        :param debug: Debug mode
        :returns: None
        """

        if self.arguments.working_dir == 'current directory':
            self.arguments.working_dir = os.getcwd()

        try:
            self.docker_cmd.docker_deploy_compss(self.arguments.working_dir,
                                self.arguments.image,
                                self.arguments.restart)

            master_ip = self.docker_cmd.docker_exec_in_daemon("hostname -i", return_output=True)
            self.env_add_conf({'master_ip': master_ip})
        except:
            traceback.print_exc()
            print("ERROR: Docker deployment failed")
            self.env_remove(env_id=self.arguments.name)


    def update(self):
        """ Deploys COMPSs infrastructure in Docker

        :param arguments: Command line arguments
        :param debug: Debug mode
        :returns: None
        """
        if self.debug:
            print("Updating...")
        self.docker_cmd.docker_update_image()


    def exec(self):
        super().exec()
        """ Execute the given command in the running Docker image

        :param arguments: Command line arguments
        :param debug: Debug mode
        :returns: None
        """
        if self.debug:
            print("Executing...")
            print("Parameters:")
            print("\t- Command: " + self.arguments.command)
            print("\t- self.Arguments: " + str(self.arguments.argument))
        command = ' '.join(self.arguments.exec_cmd)
        self.docker_cmd.docker_exec_in_daemon(command)


    def run(self):
        super().exec()
        """ Run the given command in the COMPSs infrastructure at docker

        :param arguments: Command line arguments
        :param debug: Debug mode
        :returns: None
        """
        if self.debug:
            print("Running...")
            print("Parameters:")
            print("\t- Application: " + self.arguments.application)
            print("\t- self.Arguments: " + str(self.arguments.argument))

        app_args = self.arguments.rest_args

        if '--project' not in app_args:
            app_args.insert(0, '--project=/project.xml ')
        if '--resources' not in app_args:
            app_args.insert(0, '--resources=/resources.xml ')
        if '--master_name' not in app_args:
            app_args.insert(0, f"--master_name={self.env_conf['master_ip']} ")
        if '--base_log_dir' not in app_args:
            app_args.insert(0, '--base_log_dir=/home/user ')

        command = "runcompss " + ' '.join(app_args)

        self.docker_cmd.docker_exec_in_daemon(command)

        self.docker_cmd.docker_exec_in_daemon('cp -a /home/user/.COMPSs/. /root/.COMPSs/')


    def monitor(self):
        """ Starts or stops the monitor in the COMPSs infrastructure at docker

        :param arguments: Command line arguments
        :param debug: Debug mode
        :returns: None
        """
        if self.debug:
            print("Monitoring...")
            print("Parameters:")
            print("\t- Option: " + self.arguments.option)
        if self.arguments.option == "start":
            self.docker_cmd.docker_start_monitoring()
        elif self.arguments.option == "stop":
            self.docker_cmd.docker_stop_monitoring()
        else:
            raise Exception("Unexpected monitor option: " + self.arguments.option)


    def jupyter(self):
        """ Starts jupyter in the COMPSs infrastructure at docker

        :param arguments: Command line arguments
        :param debug: Debug mode
        :returns: None
        """

        self.docker_cmd.docker_exec_in_daemon('pkill jupyter')
        
        arguments = " ".join(self.arguments.rest_args)
        working_dir = self.env_conf['working_dir']
        for arg in self.arguments.rest_args:
            dir = working_dir + '/' + arg
            if os.path.isdir(dir):
                print(f"Opening jupyter server in `{dir}`")
                break

        jupyter_cmd = "jupyter-notebook " + \
                arguments + " " + \
                f"--ip={self.env_conf['master_ip']} " + \
                "--allow-root " + \
                "--NotebookApp.token="
        self.docker_cmd.docker_exec_in_daemon(jupyter_cmd)

        if self.docker_cmd.exists():
            self.docker_cmd.docker_exec_in_daemon('pkill jupyter')


    def gengraph(self):
        """ Converts the given task dependency graph (dot) into pdf
        using the docker image

        :param arguments: Command line arguments
        :param debug: Debug mode
        :returns: None
        """
        dot_path = self.arguments.dot_file
        
        command = "compss_gengraph " + dot_path
        self.docker_cmd.docker_exec_in_daemon(command)

    def app(self):
        print("ERROR: Wrong Environment! Try using a `cluster` environment")
        exit(1)

    def components(self):
        """ Lists/add/remove workers in the COMPSs infrastructure at docker

        :param arguments: Command line arguments
        :param debug: Debug mode
        :returns: None
        """
        if self.debug:
            print("Components: " + self.arguments.components)

        if self.arguments.components == "list":
            if self.debug:
                print("Listing components...")
            self.docker_cmd.docker_components(self.arguments.components)
        if self.arguments.components == "add":
            if self.debug:
                print("Adding components: " + str(self.arguments.worker))
            self.docker_cmd.docker_components(self.arguments.components,
                            self.arguments.add,
                            self.arguments.worker)
        if self.arguments.components == "remove":
            if self.debug:
                print("Removing components:" + str(self.arguments.worker))
            self.docker_cmd.docker_components(self.arguments.components,
                            self.arguments.remove,
                            self.arguments.worker)

    def env_remove(self, env_id=None):
        if self.docker_cmd.exists():
            self.docker_cmd.docker_exec_in_daemon('rm -rf .COMPSs')
            self.docker_cmd.docker_kill_compss()
        super().env_remove(eid=env_id)

    def job(self):
        print("ERROR: Wrong Environment! Try using a `cluster` environment")
        exit(1)


#!/usr/bin/python
#
#  Copyright 2002-2021 Barcelona Supercomputing Center (www.bsc.es)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
This script is aimed at cancelling a running PyCOMPSs job with Jupyter notebook.
"""

import os
import sys
from commons import VERBOSE
from commons import SUCCESS_KEYWORD
from commons import ERROR_KEYWORD
from commons import setup_supercomputer_configuration
from commons import update_command
from commons import command_runner
from commons import is_notebook_job
from commons import not_a_notebook


def cancel():
    """
    Main cancel function.
    Gets the cancel command from the environment variable (defined in supercomputer cfgs)
    and includes the job identifiers to be cancelled.
    :return: None
    """
    if VERBOSE:
        print("Cancelling PyCOMPSs interactive job...")

    # Get command line arguments
    job_ids = sys.argv[1:]

    # Load the Supercomputer configuration to get the appropriate cancel command
    setup_supercomputer_configuration()

    success = True
    # There might be more than one to cancel
    for job_id in job_ids:
        if VERBOSE:
            print(" - Job: " + str(job_id))

        # Get the command to cancel the job
        raw_cancel_command = os.environ['QUEUE_JOB_CANCEL_CMD']
        cancel_command = update_command(raw_cancel_command, job_id)

        # Cancel the job
        return_code, _, _ = command_runner(cancel_command)

        if return_code != 0:
            success = False

    if VERBOSE:
        print("Finished cancellation.")

    # Print to notify the cancel result
    if success:
        print(SUCCESS_KEYWORD)
    else:
        print(ERROR_KEYWORD)
        exit(1)


if __name__ == '__main__':
    cancel()



import os
import getpass

# --------------------------------------------------
# PROGRAM
# --------------------------------------------------
VERSION = '0.1.0'

# --------------------------------------------------
# BACKEND SCRIPTS
# --------------------------------------------------
INTERPRETER = 'python'
SUBMIT_SCRIPT = 'submit.py'
STATUS_SCRIPT = 'status.py'
INFO_SCRIPT = 'info.py'
FIND_SCRIPT = 'find.py'
CANCEL_SCRIPT = 'cancel.py'

# --------------------------------------------------
# COMMON TAGS (backend and client)
# CAUTION! These tags must match with the ones
#          defined in backend commons.py
# --------------------------------------------------
DECODING_FORMAT = 'utf-8'
SUCCESS_KEYWORD = 'SUCCESS'
NOT_RUNNING_KEYWORD = 'NOT_RUNING'
ERROR_KEYWORD = 'ERROR'
DISABLED_VALUE = 'undefined'

# --------------------------------------------------
# ARGUMENTS
# --------------------------------------------------

LOG_LEVEL_DEBUG = 'debug'
LOG_LEVEL_INFO = 'info'
LOG_LEVEL_OFF = 'off'
DEFAULT_VERBOSE = False
DEFAULT_CONNECTIVITY_CHECK = True

DEFAULT_USER_NAME = str(getpass.getuser())

# Matches the same key as dest in arguments parsers
DEFAULT_PROJECT = {'exec_time': 5,  # 5 minute
                   'job_name': 'Jupyter',
                   'num_nodes': 2,
                   'qos': 'debug',
                   'log_level': LOG_LEVEL_OFF,
                   'tracing': False,
                   'classpath': '.',
                   'pythonpath': '.',
                   'notebook_path': DISABLED_VALUE,  # if not defined, use remote HOME
                   'storage_home': DISABLED_VALUE,
                   'storage_props': DISABLED_VALUE,
                   'storage': 'None'}

DEFAULT_CREDENTIALS = {'user_name': DEFAULT_USER_NAME,
                       'password': False,
                       'supercomputer': 'mn2.bsc.es',
                       'session': DISABLED_VALUE,
                       'port_forward': '3333',
                       'web_browser': 'firefox',
                       'no_web_browser': False}

# --------------------------------------------------
# SCRIPT TOOLS
# --------------------------------------------------
DEFAULT_SSH = 'ssh'  # When password provided, use sshpass -e before ssh
DEFAULT_SSHPASS = 'sshpass -e ' + DEFAULT_SSH
DEFAULT_SSH_WINDOWS = 'plink -batch'

# --------------------------------------------------
# SCRIPT VARIABLES
# --------------------------------------------------
CONNECTION_URL = "http://localhost:8888/?token="

# --------------------------------------------------
# INFO CONSTANTS DECLARATION
# --------------------------------------------------
INFO_CONNECTION_ESTABLISHED = "Connection established. Please use the following URL to connect to the job."

# --------------------------------------------------
# WARNING CONSTANTS DECLARATION
# --------------------------------------------------
WARNING_USER_NAME_NOT_PROVIDED = "Username not provided. Using: " + DEFAULT_USER_NAME
WARNING_NOTEBOOK_NOT_RUNNING = "The notebook is not running!"
WARNING_NO_BROWSER = "Unexpected no_web_broser value found.\nPlease use True or False. Now using False."

# --------------------------------------------------
# ERROR CONSTANTS DECLARATION
# --------------------------------------------------
ERROR_UNEXPECTED_PARAMETER = "Unexpected parameter in project file: "
ERROR_UNRECOGNIZED_ACTION = "Unrecognized action: "
ERROR_SESSION_NOT_PROVIDED = "Session not provided. Please provide one to connect."
ERROR_CONNECTING = "Could not connect to the supercomputer.\n" \
                   "Please check the connectivity, user name and password."
ERROR_COMPSS_NOT_DEFINED = "COMPSs is not available in the supercomputer.\n" \
                           "Please check that you have properly configured the module load in your .bashrc."
ERROR_SUBMITTING_JOB = "Could not submit the job to the supercomputer.\n" \
                       "Please check that you are able to submit jobs to the supercomputer."
ERROR_STATUS_JOB = "Could not check the status of the job."
ERROR_INFO_JOB = "Could not get the information of the job."
ERROR_STORAGE_PROPS = "--storage_props flag not defined."
ERROR_UNSUPPORTED_STORAGE_SHORTCUT = "Non supported external storage shortcut."
ERROR_JUPYTER_SERVER = "The notebook server is not running."
ERROR_BROWSER = "Could not open the browser. Please check that you have defined properly the browser to use."
ERROR_NO_BROWSER = "Wrong no_web_browser flag type!"
ERROR_CANCELLING_JOB = "Could not cancel the job. Please, cancel it manually."
ERROR_FORWARD_PORT = "Could not forward the port. Please check that you have defined properly the port to forward."

# --------------------------------------------------
# FUNCTIONS
# --------------------------------------------------

def is_windows():
    """ Checks if running in windows """
    return os.name == 'nt'



"""
This file contains the actions supported by pycompss-cli.
They are invoked from cli/pycompss.py and uses core/cmd.py.
"""

from pycompss_cli.core.remote.cmd import remote_app_deploy, remote_app_remove, remote_deploy_compss
from pycompss_cli.core.remote.cmd import remote_submit_job
from pycompss_cli.core.remote.cmd import remote_list_job
from pycompss_cli.core.remote.cmd import remote_cancel_job
from pycompss_cli.core.remote.cmd import remote_run_app
from pycompss_cli.core.remote.cmd import remote_exec_app
from pycompss_cli.core.remote.cmd import remote_list_apps
from pycompss_cli.core.remote.cmd import remote_get_home
from pycompss_cli.core.remote.cmd import remote_env_remove
from pycompss_cli.core.actions import Actions
from pycompss_cli.core import utils
from pycompss_cli.core.remote.interactive_sc import core, defaults
import os, json, time, datetime, traceback, re
from collections import defaultdict

class RemoteActions(Actions):
    def __init__(self, arguments, debug=False, env_conf=None) -> None:
        super().__init__(arguments, debug=debug, env_conf=env_conf)

        self.apps = None
        self.past_jobs = defaultdict(dict)
        if self.env_conf and os.path.isfile(self.env_conf['env_path'] + '/jobs.json'):
            with open(self.env_conf['env_path'] + '/jobs.json') as f:
                self.past_jobs = defaultdict(dict, json.load(f))

    def get_apps(self, env_id=None):
        if self.apps is not None:
            return self.apps

        if env_id is not None:
            if 'remote_home' not in self.env_conf:
                return []
            return remote_list_apps(env_id, self.env_conf['login'], self.env_conf['remote_home'])
        
        self.apps = remote_list_apps(self.env_conf['name'], self.env_conf['login'], self.env_conf['remote_home'])
        return self.apps

    def init(self):
        """ Deploys COMPSs infrastructure on remote|cluster env

        :param arguments: Command line arguments
        :param debug: Debug mode
        :returns: None
        """
        super().init()

        if self.arguments.modules is None:
            self.arguments.modules = ['COMPSs']
        elif len(self.arguments.modules) == 1 and os.path.isfile(self.arguments.modules[0]):
            self.arguments.modules = self.arguments.modules[0]
        elif 'COMPSs' not in [m[:len('COMPSs')] for m in self.arguments.modules]:
            self.arguments.modules.append('COMPSs')
        
        print('Deploying environment...')

        env_id = self.arguments.name
        
        try:
            remote_deploy_compss(env_id, self.arguments.login, self.arguments.modules)

            remote_home_path = remote_get_home(self.arguments.login)

            self.env_add_conf({'remote_home':  remote_home_path})
        except:
            traceback.print_exc()
            
            print("ERROR: Cluster({}) deployment failed".format(env_id))
            self.env_remove(env_id)

    def run(self):
        app_name = self.arguments.app_name
        if not app_name:
            print(f"ERROR: Application ID argument (-app) is required for executing runcompss in cluster")
            exit(1)

        if app_name not in self.get_apps():
            print(f"ERROR: Application `{app_name}` not found")
            exit(1)

        app_args = self.arguments.rest_args
        command = "runcompss " + ' '.join(app_args)
        login_info = self.env_conf['login']
        env_id = self.env_conf['name']
        remote_dir = self.env_conf['remote_home'] + f'/.COMPSsApps/{env_id}/{app_name}'
        env_name = self.env_conf['name']
        modules = self.__get_modules()

        print(remote_run_app(remote_dir, login_info, env_name, command, modules))

    def __get_modules(self):
        with open(self.env_conf['env_path'] + '/modules.sh', 'r') as mod_file:
            return mod_file.read().strip().split('\n')

    def job(self):
        super().job()

        action_name = self.arguments.job
        action_name = utils.get_object_method_by_name(self, 'job_' + action_name, include_in_name=True)
        getattr(self, action_name)()
                
    def job_submit(self):
        super().job_submit()

        app_name = self.arguments.app_name
        
        if not app_name:
            print(f"ERROR: Application ID argument (-app) is required for executing runcompss in cluster")
            exit(1)

        if app_name not in self.get_apps():
            print(f"ERROR: Application {app_name} not found")
            exit(1)

        login_info = self.env_conf['login']
        env_id = self.env_conf['name']
        remote_dir = self.env_conf['remote_home'] + f'/.COMPSsApps/{env_id}/{app_name}'

        if len(self.arguments.rest_args) == 1 and os.path.isfile(self.arguments.rest_args[0]):
            with open(self.arguments.rest_args[0], 'r') as args_file:
                app_args = args_file.read().strip().replace('\n', ' ')
        else:
            app_args = " ".join(self.arguments.rest_args)

        app_args = app_args.replace('{COMPS_APP_PATH}', remote_dir)

        app_args = f'--pythonpath={remote_dir} ' + app_args
        app_args = f'--classpath={remote_dir} ' + app_args
        app_args = f'--appdir={remote_dir} ' + app_args

        modules = self.__get_modules()
        env_vars = [item for sublist in self.arguments.env_var for item in sublist]
        if 'COMPSS_PYTHON_VERSION' not in ''.join(env_vars):
            env_vars = ['COMPSS_PYTHON_VERSION=3'] + env_vars
        
        job_id = remote_submit_job(login_info, remote_dir, app_args, modules, envars=env_vars)

        if self.arguments.verbose:
            print('\t-\tenvars:', env_vars)
            print('\t-\tenqueue_compss:', app_args)

        self.past_jobs[job_id] = {
            'app_name': app_name,
            'env_vars': '; '.join(env_vars) if env_vars else 'None',
            'enqueue_args': app_args,
            'timestamp': str(datetime.datetime.utcnow())[:-7] + ' UTC'
        }

        with open(self.env_conf['env_path'] + '/jobs.json', 'w') as f:
            json.dump(self.past_jobs, f)

    def job_history(self):
        job_id = self.arguments.job_id
        if job_id:
            app_jobs = self.past_jobs[job_id]
            print('\tApp name:', app_jobs['app_name'])
            print('\tSubmit time:', app_jobs['timestamp'])
            print('\tEnvironment Variables:', app_jobs['env_vars'])
            print('\tEnqueue Args:', app_jobs['enqueue_args'])
            print()
        else:
            col_names = ['JobIDs', 'AppName', 'Status']
            rows = []
            for job_id, app_job in self.past_jobs.items():
                if 'status' not in app_job:
                    job_status = self.job_status(job_id)
                else:
                    job_status = app_job['status']
                    if 'COMPLETED' not in job_status:
                        job_status = self.job_status(job_id)
                if job_status is None:
                    job_status = 'Unknown'
                rows.append([job_id, app_job['app_name'], f'\t{job_status}'])
            utils.table_print(col_names, rows)

    def job_list(self):
        login_info = self.env_conf['login']
        modules = self.__get_modules()
        remote_list_job(login_info, modules)

    def job_cancel(self):
        login_info = self.env_conf['login']
        jobid = self.arguments.job_id
        modules = self.__get_modules()
        remote_cancel_job(login_info, jobid, modules)

    def job_status(self, jid=None):
        login_info = self.env_conf['login']
        job_id = self.arguments.job_id if jid is None else jid
        modules = self.__get_modules()
        scripts_path = self.env_conf['remote_home'] + '/.COMPSs/job_scripts'
        job_status = core.job_status(scripts_path, job_id, login_info, modules)
        if job_status == 'ERROR' and job_id in self.past_jobs:
            app_name = self.past_jobs[job_id]['app_name']
            env_id = self.env_conf['name']
            app_path = self.env_conf['remote_home'] + f'/.COMPSsApps/{env_id}/{app_name}'
            cmd = f'grep -iF "error" {app_path}/compss-{job_id}.err'
            status = 'ERROR' if remote_exec_app(login_info, cmd) else 'SUCCESS'
            job_status = 'COMPLETED:' + status
        
        if jid is None:
            print(job_status)

        if job_id in self.past_jobs:
            self.past_jobs[job_id]['status'] = job_status
            with open(self.env_conf['env_path'] + '/jobs.json', 'w') as f:
                json.dump(self.past_jobs, f)

        return job_status

    def app(self):
        if not self.arguments.app:
            self.arguments.func()
            exit(1)

        action_name = self.arguments.app
        action_name = utils.get_object_method_by_name(self, 'app_' + action_name, include_in_name=True)
        getattr(self, action_name)()

    def app_deploy(self):
        if self.arguments.local_source == 'current directory':
            self.arguments.local_source = os.getcwd()
        else:
            if not os.path.isdir(self.arguments.local_source):
                print(f"ERROR: Local source directory {self.arguments.local_source} does not exist")
                exit(1)

        app_name = self.arguments.app_name
        if app_name in self.get_apps():
            print(f'ERROR: There is already another application named `{app_name}`')
            exit(1)

        # if self.arguments.remote_dir:
        #     self.arguments.remote_dir = self.arguments.remote_dir.replace('{COMPSS_REMOTE_HOME}', )

        env_id = self.env_conf['name']
        app_dir = self.env_conf['remote_home'] + f'/.COMPSsApps/{env_id}/{app_name}'

        remote_app_deploy(app_dir, self.env_conf['login'], self.arguments.local_source, self.arguments.remote_dir)


    def app_remove(self):
        app_names = self.arguments.app_name
        for app_name in app_names:
            if app_name in self.get_apps():
                login_info = self.env_conf['login']
                env_id = self.env_conf['name']
                app_dir = self.env_conf['remote_home'] + f'/.COMPSsApps/{env_id}/{app_name}'
                remote_app_remove(login_info, app_dir)
                print(f'Application `{app_name}` removed successfully')
            else:
                print('ERROR: Application not found')
                exit(1)

    def app_list(self):
        apps = self.get_apps()

        if not apps:
            print('INFO: There are no applications binded to this environment yet')
            print('       Try deploying an application to the cluster')
            exit(1)

        utils.table_print(['Name'], [[a] for a in apps])

    def exec(self):
        login_info = self.env_conf['login']
        command = ' '.join(self.arguments.exec_cmd)
        remote_exec_app(login_info, command)

    def env_remove(self, eid=None):
        env_id = self.arguments.env_id if eid is None else eid
        env_apps = self.get_apps(env_id=env_id)
        if len(env_apps) > 0:
            print('WARNING: There are still applications binded to this environment')
            answer = 'y'
            if not self.arguments.force:
                answer = input('Do you want to delete this environment and all the applications? (y/N) ')
            if answer == 'Y' or answer == 'y' or answer == 'yes':
                login_info = self.env_conf['login']
                remote_env_remove(login_info, env_id, env_apps)
                super().env_remove(eid=eid)
        else:
            super().env_remove(eid=eid)

    def components(self):
        print('ERROR: Not Implemented Yet')
        exit(1)

    def gengraph(self):
        print('ERROR: Not Implemented Yet')
        exit(1)

    def monitor(self):
        print('ERROR: Not Implemented Yet')
        exit(1)

    def jupyter(self):
        app_name = self.arguments.app_name
        if not app_name:
            print(f"ERROR: Application ID argument (-app) is required for starting jupyter on cluster")
            exit(1)

        if app_name not in self.get_apps():
            print(f"ERROR: Application `{app_name}` not found")
            exit(1)
    

        login_info = self.env_conf['login']
        env_id = self.env_conf['name']
        remote_dir = self.env_conf['remote_home'] + f'/.COMPSsApps/{env_id}/{app_name}'
        job_name = app_name + '-PyCOMPSsInteractive'
        app_args = ['--jupyter_notebook='+remote_dir, 
                    '--job_name='+job_name, 
                    '--lang=python',
                    f'--master_working_dir={remote_dir}',
                    f'--pythonpath={remote_dir}']
        app_args += self.arguments.rest_args

        port = '8888'
        for arg in app_args:
            if '--port=' in arg:
                port = arg.split('=')[1]
                app_args.remove(arg)
                break

        app_args = ' '.join(app_args)
        modules = self.__get_modules()
        envars = []
        if re.search(r'mn\d\.bsc\.es', self.env_conf['login']) is not None:
            envars.append('COMPSS_PYTHON_VERSION=3-jupyter')

        job_id = remote_submit_job(login_info, remote_dir, app_args, modules, envars)

        scripts_path = self.env_conf['remote_home'] + '/.COMPSs/job_scripts'

        print('Waiting for jupyter to start...')
        jupyter_job_status = defaults.NOT_RUNNING_KEYWORD
        try:
            while jupyter_job_status != 'RUNNING':
                jupyter_job_status = core.job_status(scripts_path, job_id, login_info, modules)
        except:
            print('ERROR while waiting for jupyter to start')
            core.cancel_job(scripts_path, [job_id], login_info, modules)
            exit(1)
        else:
            print('Jupyter started')

            print('Connecting to jupyter server...')
            time.sleep(5)

            core.connect_job(scripts_path, job_id, login_info, modules, remote_dir, port_forward=port, web_browser=None)
            
            core.cancel_job(scripts_path, [job_id], login_info, modules)


import os
from re import sub
from sys import stdout
import tempfile
from shutil import copyfile
import subprocess
import tempfile
import pycompss_cli.core.utils as utils

# ################ #
# GLOBAL VARIABLES #
# ################ #

# ############# #
# API FUNCTIONS #
# ############# #

def remote_deploy_compss(env_id: str, login_info: str, modules, envars=[]) -> None:
    """ Deploy environment COMPSs in remote cluster.
    It stops any existing one since it can not coexist with itself.

    :param working_dir: Given working directory
    :param image: Given docker image
    :param restart: Force stop the existing and start a new one.
    :returns: None
    """

    subprocess.run(f"ssh {login_info} 'mkdir -p ~/.COMPSs'", shell=True)

    if isinstance(modules, str):
        modules_file = modules
    else:
        tmp_modules = tempfile.NamedTemporaryFile(delete=False)
        env_vars = '\n'.join([f'export {var}' for var in envars])
        mod_script = '\n'.join([f'module load {m}' for m in modules])
        tmp_modules.write((env_vars + '\n').encode())
        tmp_modules.write((mod_script + '\n').encode())
        tmp_modules.close()
        modules_file = tmp_modules.name

    copyfile(modules_file, f'{os.path.expanduser("~")}/.COMPSs/envs/{env_id}/modules.sh')

    if not isinstance(modules, str):
        os.unlink(tmp_modules.name)

    path = os.path.abspath(__file__)
    local_job_scripts_dir = os.path.dirname(path) + '/job_scripts/'
    subprocess.run(f"scp -r {local_job_scripts_dir} '{login_info}:~/.COMPSs/'", shell=True, stdout=subprocess.PIPE)
    

def remote_app_deploy(app_dir: str, login_info: str, local_source: str, remote_dest_dir: str = None):
    print('Deploying app...')
    utils.ssh_run_commands(login_info, [f'mkdir -p {app_dir}'])

    cmd_copy_files = f"scp -r {local_source}/* {login_info}:'{app_dir}/'"

    if os.path.isfile(local_source):
        cmd_copy_files = cmd_copy_files.replace('/*', '')

    subprocess.run(cmd_copy_files, shell=True)

    if remote_dest_dir:
        print('App deployed to', remote_dest_dir)
        utils.ssh_run_commands(login_info, [
            f'echo {remote_dest_dir} > {app_dir}/.compss',
            f'ln -s {app_dir} {remote_dest_dir}'
        ])

def remote_app_remove(login_info: str, app_dir: str):
    utils.ssh_run_commands(login_info, [
            f'cat {app_dir}/.compss | xargs rm',
            f'rm -rf {app_dir}'
        ])

def remote_run_app(remote_dir: str, login_info: str, env_name: str, command: str, modules):
    commands = [
        f'cd {remote_dir}',
        *modules,
        command
    ]
    return utils.ssh_run_commands(login_info, commands)

def remote_submit_job(login_info: str, remote_dir: str, app_args: str, modules, envars=None) -> None:
    """ Execute the given command in the cluster COMPSs environment.

    :param cmd: Command to execute.
    :returns: The execution stdout.
    """

    enqueue_debug = '-d' if utils.is_debug() else ''

    commands = [
        f'cd {remote_dir}',
        *modules,
        f'enqueue_compss {enqueue_debug} {app_args}'
    ]

    if envars:
        commands = [f'export {var}' for var in envars] + commands

    if utils.is_debug():
        print('********* DEBUG *********')
        print('Remote submit job commands:')
        for cmd in commands:
            print('\t', '->', cmd)
        print('***************************')

    stdout = utils.ssh_run_commands(login_info, commands)
    job_id = stdout.strip().split('\n')[-1].split(' ')[-1]
    print('Job submitted:', job_id)
    return job_id

def remote_get_home(login_info: str):
    return utils.ssh_run_commands(login_info, ['echo $HOME']).strip()

def remote_list_apps(env_id: str, login_info: str, remote_home: str):
    commands = [
        f'[ ! -d "{remote_home}/.COMPSsApps/{env_id}" ] && echo "NO_APPS"',
        f'ls ~/.COMPSsApps/{env_id}/',
    ]

    stdout = utils.ssh_run_commands(login_info, commands).strip()
    apps = stdout.split('\n')
    if 'NO_APPS' in stdout or (len(apps) == 1 and apps[0] == ''):
        return []

    return apps

def remote_env_remove(login_info, env_id, env_apps):
    for app_name in env_apps:
        print(f'Deleting {app_name}...')
        remote_dir = f'~/.COMPSsApps/{env_id}/{app_name}'
        remote_app_remove(login_info, remote_dir)

    utils.ssh_run_commands(login_info, [f'rm -rf ~/.COMPSsApps/{env_id}',])

def remote_list_job(login_info: str, modules):

    commands = [
        *modules,
        f'python3 ~/.COMPSs/job_scripts/find.py',
    ]

    stdout = utils.ssh_run_commands(login_info, commands).strip()
    if stdout != 'SUCCESS':
        print(stdout)
    else:
        print('No jobs found')

def remote_cancel_job(login_info: str, job_id: str, modules):
    commands = [
        *modules,
        f'python3 ~/.COMPSs/job_scripts/cancel.py {job_id}',
    ]

    stdout = utils.ssh_run_commands(login_info, commands).strip()
    print(stdout)

def remote_exec_app(login_info: str, exec_cmd: str):
   return utils.ssh_run_commands(login_info, [exec_cmd]).strip()


#!/usr/bin/env python3

import os
import pycompss_cli.cli.pycompss as pycompss


def main():
    # Force to use the dislib image by default
    os.environ['DEFAULT_DISLIB_DOCKER_IMAGE'] = 'bscwdc/dislib:latest'
    pycompss.main()

if __name__ == "__main__":
    main()



#!/usr/bin/python
#
#  Copyright 2002-2021 Barcelona Supercomputing Center (www.bsc.es)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
This script is aimed at checking the information of a running PyCOMPSs jobs
with Jupyter notebook.
"""

import os
import sys
import time
from commons import SUCCESS_KEYWORD
from commons import ERROR_KEYWORD
from commons import NOT_RUNNING_KEYWORD
from commons import setup_supercomputer_configuration
from commons import get_jupyter_environment_variables
from commons import update_command
from commons import command_runner
from commons import command_runner_shell
from commons import is_notebook_job
from commons import not_a_notebook
from pycompss_cli.core import utils


def info():
    """
    Main info function.
    Gets the information command from the jupyter server (master node and url)
    and prints it through stdout. So that the client can get it, parse and
    connect to the running notebook.
    :return: None
    """
    if utils.is_debug():
        print("Checking information of PyCOMPSs interactive job...")

    # Get command line arguments
    job_id = sys.argv[1]
    app_path = sys.argv[2]

    if utils.is_debug():
        print(" - Job: " + str(job_id))

    # Load the Supercomputer configuration to get the appropriate status command
    setup_supercomputer_configuration()
    
    # Get the list of nodes
    raw_nodes_command = os.environ['QUEUE_JOB_NODES_CMD']
    nodes_command = update_command(raw_nodes_command, job_id)
    return_code, nodes, _ = command_runner(nodes_command)
    if return_code != 0:
        print(ERROR_KEYWORD)
        exit(1)

    # Look for the master node
    if nodes.strip():
        nodes_expansor = os.environ['HOSTLIST_CMD'] + ' ' + nodes
        _, expanded_nodes, _ = command_runner(nodes_expansor.split())
        expanded_nodes = expanded_nodes.splitlines()
        master = expanded_nodes[0]

        if utils.is_debug():
            print(" - Nodes: " + str(expanded_nodes))
            print(" - Found master: " + str(master))

        jpy_server_out = ''
        timeout = time.time() + 60*2
        jup_out_file = f'{app_path}/compss-{job_id}.err'
        while 'The Jupyter Notebook is running at' not in jpy_server_out:
            if time.time() > timeout:
                print(ERROR_KEYWORD)
                exit(1)
            if os.path.isfile(jup_out_file):
                with open(jup_out_file, 'r') as of:
                    jpy_server_out = of.read()

        if utils.is_debug():
            print("Finished checking the information.")

        # Print to notify the info result
        print(SUCCESS_KEYWORD)
        print("MASTER:" + str(master))
        print("SERVER: " + str(jpy_server_out.replace('\n', ' ')))
    else:
        # Print to provide information to the client
        print(NOT_RUNNING_KEYWORD)  # The notebook is not ready yet
        exit(1)


if __name__ == '__main__':
    info()



#!/usr/bin/python
#
#  Copyright 2002-2021 Barcelona Supercomputing Center (www.bsc.es)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
This file contains all common functions that the backend scripts that provide
support for PyCOMPSs and Jupyter within Supercomputers.
"""

import os
import shlex
import subprocess

VERBOSE = False   # General Boolean to print detailed information through stdout - For debugging purposes.
DECODING_FORMAT = 'utf-8'
SUCCESS_KEYWORD = 'SUCCESS'
NOT_RUNNING_KEYWORD = 'NOT_RUNING'
ERROR_KEYWORD = 'ERROR'
DISABLED_VALUE = 'undefined'
JOB_NAME_KEYWORD = '-PyCOMPSsInteractive'


def command_runner(cmd, exception=True, cwd=None):
    """
    Run the command defined in the cmd list (for single commands).
    Decodes the stdout and stderr following the DECODING_FORMAT.
    :param cmd: Command to execute as list.
    :param exception: Throw exception if failed. Otherwise, the caller will handle the error.
    :param cwd: Directory where to execute the command
    :return: return code, stdout, stderr
    """
    if VERBOSE:
        print("Executing command: " + ' '.join(cmd))
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
    stdout, stderr = p.communicate()   # blocks until cmd is done
    return_code = p.returncode
    stdout = stdout.decode(DECODING_FORMAT)
    stderr = stderr.decode(DECODING_FORMAT)
    if exception and return_code != 0:
        _raise_command_exception(cmd, return_code, stdout, stderr)
    return return_code, stdout, stderr


def command_runner_shell(cmd, exception=True, cwd=None):
    """
    Run the command defined in the cmd string.
    Enables to run multiple commands at once (e.g. using &&).
    Decodes the stdout and stderr following the DECODING_FORMAT.
    :param cmd: Command to execute as string.
    :param exception: Throw exception if failed. Otherwise, the caller will handle the error.
    :param cwd: Directory where to execute the command
    :return: return code, stdout, stderr
    """
    if VERBOSE:
        print("Executing command: " + ' '.join(cmd))
    current_env = os.environ.copy()
    p = subprocess.Popen(cmd,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         cwd=cwd,
                         env=current_env,
                         shell=True)
    stdout, stderr = p.communicate()   # blocks until cmd is done
    return_code = p.returncode
    stdout = stdout.decode(DECODING_FORMAT)
    stderr = stderr.decode(DECODING_FORMAT)
    if exception and return_code != 0:
        _raise_command_exception(cmd, return_code, stdout, stderr)
    return return_code, stdout, stderr

def get_jupyter_environment_variables():
    """
    Get the jupyter environment variables as dictionary.
    # export JUPYTER_RUNTIME_DIR=$HOME/jupyter-runtime
    :return: Dictionary with the required environment variables.
    """
    return {"JUPYTER_RUNTIME_DIR": os.environ["HOME"] + "/jupyter-runtime"}

def setup_jupyter_environment_only():
    """
    Setup the jupyter environment variables.
    :return: None
    """
    if VERBOSE:
        print("Setting up the jupyter environment...")
    include = get_jupyter_environment_variables()
    _export_environment_variables(None, include)

def setup_supercomputer_configuration():
    """
    Setup the supercomputer configuration.
    Reads the supercomputer cfg and queuing system cfg and exports
    the variables defined in them within the current environment.
    This is useful to get the appropriate commands for the submission,
    cancel, status of jobs, etc. which depends on the specific supercomputer
    relying on the COMPSs configuration.
    Alternatively, it is possible to include a keyword in the command for
    the later substitution from the python code before using it (e.g.
    QUEUE_JOB_NAME_CMD - which is used multiple times).
    :return: None
    """
    if VERBOSE:
        print("Setting up the supercomputer configuration...")
    # Include the jupyter environment variables
    include = get_jupyter_environment_variables()
    # Get the scripts directory
    sc_cfg_name = 'default.cfg'
    sc_cfg = os.path.join(os.environ["COMPSS_HOME"],
                          'Runtime', 'scripts', 'queues', 'supercomputers', sc_cfg_name)
    if VERBOSE:
        print("* Loading SC cfg file")
    _export_environment_variables(sc_cfg, include)
    # Export the environment variables
    qs_cfg_name = str(os.environ['QUEUE_SYSTEM'].strip()) + '.cfg'
    qs_cfg = os.path.join(os.environ["COMPSS_HOME"],
                          'Runtime', 'scripts', 'queues', 'queue_systems', qs_cfg_name)
    if VERBOSE:
        print("* Loading QS cfg file")
    _export_environment_variables(qs_cfg, include)


def is_notebook_job(job_id):
    """
    Checks if the given job id is running a PyCOMPSs notebook.
    To this end, checks the job name to see if it matches the JOB_NAME_KEYWORD.
    :param job_id: Job id to check
    :return: True if is a notebook. False on the contrary
    """
    name = get_job_name(job_id)
    if verify_job_name(name):
        if VERBOSE:
            print("Found notebook id: " + str(job_id))
        return True
    else:
        if VERBOSE:
            print("Job " + str(job_id) + " is not a PyCOMPSs notebook job.")
        return False


def update_command(command, job_id):
    """
    Updates the given command with the necessary job_id.
    Does the replacement of %JOBID% with the job identifier
    :param command: Command to update
    :param job_id: Job identifier
    :return: The updated command as list splitter by spaces
    """
    updated_command = command.replace('%JOBID%', str(job_id)).split()
    return updated_command


def get_job_name(job_id):
    """
    Get the job name of a given job identifier.
    :param job_id: Job identifier
    :return: Job name
    """
    raw_job_name_command = os.environ['QUEUE_JOB_NAME_CMD']
    job_name_command = update_command(raw_job_name_command, str(job_id))
    _, name, _ = command_runner(job_name_command)
    return name.strip()


def verify_job_name(name):
    """
    Verifies if the name provided includes the keyword
    :param name: Name to check
    :return: Boolean
    """
    if JOB_NAME_KEYWORD in name:
        return True
    else:
        return False


def get_job_status(job_id):
    """
    Retrieves the status for the given job identifier
    :param job_id: Job identifier
    :return: The status as string, the return code
    """
    # Get the command to check the status the job
    raw_job_status_command = os.environ['QUEUE_JOB_STATUS_CMD']
    job_status_command = update_command(raw_job_status_command, job_id)
    # Check the status of the job
    return_code, stdout, stderr = command_runner(job_status_command, exception=False)
    # Get the Running tag and check if matches
    running_tag = os.environ['QUEUE_JOB_RUNNING_TAG']
    # Print to provide status to the client
    if return_code != 0:
        job_status = "CHECK FAILED"
    elif stdout.strip() == running_tag:
        job_status = "RUNNING"
    else:
        job_status = str(stdout).strip()
    return job_status, return_code


def not_a_notebook(job_id):
    """
    Prints the not a notebook job message and exit(1).
    :param job_id: Job id
    :return: None
    """
    print(ERROR_KEYWORD)
    print(" - Job Id: " + str(job_id) + " does not belong to a PyCOMPSs interactive job.")
    exit(1)


def _export_environment_variables(env_file=None, include=None):
    """
    Export the environment variables defined in "file".
    Uses bash to parse the file and gets the variables from a clean environment
    to read them and set them in the current environment.
    Includes the "include" variables in the environment if defined, so that
    the sourced configuration files can take them as inputs (e.g. job_id).
    :param env_file: File with the environment variables
    :param include: Dictionary with environment variables to consider in the
                    environment so that the cfgs can complete their contents.
    :return: None
    """
    if include:
        exports = []
        for k, v in include.items():
            exports.append('export ' + k.strip() + '=' + v.strip())
        exports = ' && '.join(exports)
        if env_file:
            command = shlex.split("env -i bash -c 'set -a && " + exports + " && source " + str(env_file) + " && env'")
        else:
            command = shlex.split("env -i bash -c 'set -a && " + exports + " && env'")
    else:
        if env_file:
            command = shlex.split("env -i bash -c 'set -a && source " + str(env_file) + " && env'")
        else:
            raise Exception("Missing at least one parameter: env_file or include")
    if VERBOSE:
        if env_file:
            print("Exporting environment variables from: " + env_file)
        if include:
            print("Exporting environment variables: " + include)
    return_code, stdout, stderr = command_runner(command)
    for line in stdout.splitlines():
        (key, _, value) = line.partition("=")
        if key != '_':
            os.environ[key] = value.strip()


def _raise_command_exception(command, return_code, stdout, stderr):
    """
    Generic exception raiser for command execution.
    :param command: Command that threw the exception
    :param return_code: Return code of the command that failed
    :param stdout: Standard output
    :param stderr: Standard error
    :return: None
    :raises: Exception
    """
    raise Exception(ERROR_KEYWORD + ": Failed execution: " + str(command)
                    + "\nRETURN CODE:" + str(return_code)
                    + "\nSTDOUT:\n" + str(stdout)
                    + "\nSTDERR:\n" + str(stderr))



import sys
import argparse


FORMATTER_CLASS = argparse.ArgumentDefaultsHelpFormatter


def local_init_parser():
    """ Parses the sys.argv.

    :returns: All arguments as namespace.
    """
    parser = argparse.ArgumentParser(formatter_class=FORMATTER_CLASS)

    # Parent parser - includes all arguments which are common to all actions
    parent_parser = argparse.ArgumentParser(add_help=False,
                                            formatter_class=FORMATTER_CLASS)
    # Action sub-parser
    subparsers = parser.add_subparsers(dest="action")
    # INIT
    parser_init = subparsers.add_parser("init",
                                        aliases=["i"],
                                        help="Initialize COMPSs locally.",
                                        parents=[parent_parser],
                                        formatter_class=FORMATTER_CLASS)
    parser_init.add_argument("-w", "--working_dir",
                             default='current directory',
                             type=str,
                             help="Working directory")

    parser_init.add_argument("-m", "--modules",
                             nargs='*',
                             help="Module list or file to load in cluster")
    
    return parser_init


import sys
import logging
import subprocess

DECODING_FORMAT = 'utf-8'


def command_runner(cmd):
    """ Run the command defined in the cmd list.

    :param cmd: Command to execute as list (list[str]).
    :returns: Exit code
    :raises Exception: Exit code != 0
    """
    print("Executing: %s" % " ".join(cmd))
    p = subprocess.Popen(cmd,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()   # blocks until cmd is done
    stdout = stdout.decode(DECODING_FORMAT)
    stderr = stderr.decode(DECODING_FORMAT)
    return_code = p.returncode
    print("Exit code: %s" % str(return_code))
    print("------------ STDOUT ------------",
          flush=True)
    print(stdout, flush=True)
    if stderr:
        print("------------ STDERR ------------",
              file=sys.stderr, flush=True)
        print(stderr, file=sys.stderr, flush=True)
    if return_code != 0:
        print("Exit code: %s != 0" % str(return_code),
              file=sys.stderr, flush=True)
    exit(return_code)



import setuptools

setuptools.setup(
    # Metadata
    name="pycompss-cli",
    version=open("VERSION.txt").read().strip(),
    description="PyCOMPSs cli",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    author="Workflows and Distributed Computing Group (WDC) - Barcelona Supercomputing Center (BSC)",
    author_email="support-compss@bsc.es",
    url="https://compss.bsc.es",

    # License
    license="Apache 2.0",

    # Build
    include_package_data=True,
    packages=setuptools.find_packages(),
    classifiers=["Development Status :: 5 - Production/Stable",
                 "Environment :: Console",
                 "Intended Audience :: Developers",
                 "Intended Audience :: Science/Research",
                 "License :: OSI Approved :: Apache Software License",
                 "Operating System :: POSIX :: Linux",
                 "Operating System :: Unix",
                 "Operating System :: MacOS",
                 "Programming Language :: Python :: 3 :: Only",
                 "Topic :: Software Development",
                 "Topic :: Scientific/Engineering",
                 "Topic :: System :: Distributed Computing",
                 "Topic :: Utilities"],
    install_requires=["setuptools"],

    # Executable
    entry_points={
        "console_scripts": [
            "compss=pycompss_cli.cli.compss:main",
            "dislib=pycompss_cli.cli.dislib:main",
            "pycompss=pycompss_cli.cli.pycompss:main",
        ],
    }
)



from pycompss_cli.core.local.actions import LocalActions
from pycompss_cli.core.docker.actions import DockerActions
from pycompss_cli.core.remote.actions import RemoteActions
from pycompss_cli.core import utils
from copy import deepcopy
import os
from pathlib import Path
import json

class ActionsDispatcher(object):
    def __init__(self) -> None:
        super().__init__()
        self.home_path = str(Path.home())

    def run_action(self, arguments, debug):
        self.__ensure_default_env()

        if 'env' in arguments and arguments.env:
            env_type = arguments.env
            env_conf = None
        else:
            envs_path = self.home_path + '/.COMPSs/envs'
            if not os.path.isdir(envs_path) or len(list(os.walk(envs_path))) == 1:
                print("There are no environments created. Try using `pycompss init`")
                exit(1)

            if arguments.action == 'environment':
                if arguments.environment and arguments.environment.startswith('r'):
                    self.__delete_envs(arguments.env_id, arguments)
                
            env_conf = utils.get_current_env_conf()
            env_type = env_conf['env']

        self.__actions_cmd = self.__getactions_cmd(env_type, arguments, debug, env_conf)

        action_name = utils.get_object_method_by_name(self.__actions_cmd, arguments.action)
        action_func = getattr(self.__actions_cmd, action_name)
        action_func()

    def __delete_envs(self, envs_ids, arguments):
        for env_id in envs_ids:
            env_type = self.__get_env_type_from_name(env_id)
            if env_type is None:
                print("ERROR: There's no environment named " + env_id)
                continue
            if env_id == 'default':
                print('ERROR: `default` environment is required and cannot be deleted')
                continue
            env_conf = utils.get_env_conf_by_name(env_id)
            env_arguments = deepcopy(arguments)
            env_arguments.env_id = env_id
            action_cmd = self.__getactions_cmd(env_type, env_arguments, env_conf=env_conf)
            action_cmd.env_remove()
        exit(0)

    def __get_env_type_from_name(self, env_name):
        envs_path = self.home_path + '/.COMPSs/envs'
        env_dir_tree = list(os.walk(envs_path))
        envs_names = env_dir_tree[0][1]
        env_dir_tree = env_dir_tree[1:]
        for i in range(len(env_dir_tree)):
            if env_name == envs_names[i]:
                return json.load(open(env_dir_tree[i][0] + '/env.json'))['env']
        return None

    def __getactions_cmd(self, env_type, arguments, debug=False, env_conf=None):
        if env_type == "local":
            return LocalActions(arguments, debug, env_conf)
        elif env_type == "docker":
            return DockerActions(arguments, debug, env_conf)
        elif env_type == "cluster":
            return RemoteActions(arguments, debug, env_conf)

    
    def __ensure_default_env(self):
        default_env = self.home_path + '/.COMPSs/envs/default'
        if not os.path.isdir(default_env):
            os.makedirs(default_env)
            open(default_env + '/current', 'a').close()
            with open(default_env + '/env.json', 'w') as def_env:
                json.dump({ 'env': 'local', 'name': 'default' }, def_env)
            with open(default_env + '/modules.sh', 'w') as mod_file:
                mod_file.write('module load COMPSS')


import os
import sys
import argparse
import subprocess
import re
import time
import signal
import getpass
from pycompss_cli.core import utils

from pycompss_cli.core.remote.interactive_sc.defaults import INTERPRETER
from pycompss_cli.core.remote.interactive_sc.defaults import SUBMIT_SCRIPT
from pycompss_cli.core.remote.interactive_sc.defaults import STATUS_SCRIPT
from pycompss_cli.core.remote.interactive_sc.defaults import INFO_SCRIPT
from pycompss_cli.core.remote.interactive_sc.defaults import FIND_SCRIPT
from pycompss_cli.core.remote.interactive_sc.defaults import CANCEL_SCRIPT

from pycompss_cli.core.remote.interactive_sc.defaults import VERSION
from pycompss_cli.core.remote.interactive_sc.defaults import DECODING_FORMAT
from pycompss_cli.core.remote.interactive_sc.defaults import SUCCESS_KEYWORD
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_KEYWORD
from pycompss_cli.core.remote.interactive_sc.defaults import NOT_RUNNING_KEYWORD
from pycompss_cli.core.remote.interactive_sc.defaults import DISABLED_VALUE

from pycompss_cli.core.remote.interactive_sc.defaults import LOG_LEVEL_DEBUG
from pycompss_cli.core.remote.interactive_sc.defaults import LOG_LEVEL_INFO
from pycompss_cli.core.remote.interactive_sc.defaults import LOG_LEVEL_OFF
from pycompss_cli.core.remote.interactive_sc.defaults import DEFAULT_VERBOSE
from pycompss_cli.core.remote.interactive_sc.defaults import DEFAULT_CONNECTIVITY_CHECK

from pycompss_cli.core.remote.interactive_sc.defaults import DEFAULT_PROJECT
from pycompss_cli.core.remote.interactive_sc.defaults import DEFAULT_CREDENTIALS

from pycompss_cli.core.remote.interactive_sc.defaults import DEFAULT_SSH
from pycompss_cli.core.remote.interactive_sc.defaults import DEFAULT_SSHPASS
from pycompss_cli.core.remote.interactive_sc.defaults import DEFAULT_SSH_WINDOWS

from pycompss_cli.core.remote.interactive_sc.defaults import CONNECTION_URL

from pycompss_cli.core.remote.interactive_sc.defaults import INFO_CONNECTION_ESTABLISHED

from pycompss_cli.core.remote.interactive_sc.defaults import WARNING_USER_NAME_NOT_PROVIDED
from pycompss_cli.core.remote.interactive_sc.defaults import WARNING_NOTEBOOK_NOT_RUNNING
from pycompss_cli.core.remote.interactive_sc.defaults import WARNING_NO_BROWSER

from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_UNEXPECTED_PARAMETER
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_UNRECOGNIZED_ACTION
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_SESSION_NOT_PROVIDED
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_CONNECTING
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_COMPSS_NOT_DEFINED
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_SUBMITTING_JOB
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_STATUS_JOB
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_INFO_JOB
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_STORAGE_PROPS
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_UNSUPPORTED_STORAGE_SHORTCUT
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_JUPYTER_SERVER
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_BROWSER
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_NO_BROWSER
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_CANCELLING_JOB
from pycompss_cli.core.remote.interactive_sc.defaults import ERROR_FORWARD_PORT

from pycompss_cli.core.remote.interactive_sc.defaults import is_windows

ALIVE_PROCESSES = []  # Needed for proper cleanup

def __signal_handler(sig, frame):
    """
    Signal handler. Acts when CTRL + C is pressed.
    Checks the global variables to see what needs to be cleaned:
        - Alive processes
        - Cancel running job
    """
    global ALIVE_PROCESSES
    
    if ALIVE_PROCESSES:
        for p in ALIVE_PROCESSES:
            p.kill()
    # # Cancel # Not cancelling here... wait for the user to cancel it explicitly.
    # # If needed, the following information needs to be global
    # global user_name
    # global supercomputer
    # global scripts_path
    # global job_id
    # global verbose
    # if job_id:
    #     if VERBOSE:
    #         print("\t - Cancelling job...")
    #     _cancel_job(user_name, supercomputer, scripts_path, job_id, verbose)
    # else:
    #     __display_warning(WARNING_JOB_CANCELLED)
    print("Finished!")
    # sys.exit(0)


def __display_error(message, return_code=None, stdout=None, stderr=None):
    """
    Display error in a common format.
    :return: None
    """
    # # Hide the client stack trace and show only the prints from remote when fails.
    # if return_code:
    #     print("Return code: " + str(return_code))
    # if stdout:
    #     print("Standard OUTPUT:")
    #     print(stdout)
    if stderr:
        print("Standard ERROR:")
        print(stderr)
    print("ERROR: " + message)
    exit(1)


def __display_warning(message):
    """
    Display warning in a common format.
    :return: None
    """
    print("WARNING: " + message)


def job_status(scripts_path, job_id, login_info, modules):
    """
    Checks the status of a job in the supercomputer.
    :param scripts_path: Remote helper scripts path
    :param job_id: Job identifier
    :return: None
    """
    cmd = [INTERPRETER,
           str(scripts_path + '/' + STATUS_SCRIPT),
           job_id]
    return_code, stdout, stderr = _command_runner(cmd, login_info, modules=modules)
    if return_code != 0:
        # __display_error(ERROR_STATUS_JOB, return_code, stdout, stderr)
        return 'ERROR'

    # Parse the output for fancy printing
    out = stdout.splitlines()
    if out[-2] == SUCCESS_KEYWORD:
        status = out[-1].split(':')[1].strip()
        if not status:
            return 'ERROR'
        return status
    return ERROR_STATUS_JOB

def job_list(scripts_path, login_info, modules):
    """
    Checks the list of available jobs in the supercomputer.
    :param scripts_path: Remote helper scripts path
    :return: None
    """
    cmd = [INTERPRETER,
           str(scripts_path + '/' + FIND_SCRIPT)]
    return_code, stdout, stderr = _command_runner(cmd, login_info, modules=modules)
    if return_code != 0:
        __display_error(ERROR_STATUS_JOB, return_code, stdout, stderr)

    # Parse the output for fancy printing
    out = stdout.splitlines()
    if out[0] == SUCCESS_KEYWORD:
        print("Available notebooks: ")
        for job_id in out[1:]:
            print(job_id)
    else:
        __display_error(ERROR_STATUS_JOB, return_code, stdout, stderr)

def connect_job(scripts_path, job_id, login_info, modules, app_path, port_forward='8888', web_browser='firefox'):
    """
    Establish the connection with an existing notebook.
    :param scripts_path: Remote helper scripts path
    :param arguments: Arguments received from command line.
    :return: None
    """
    # First register the signal (the connection will be ready until CTRL+C)
    signal.signal(signal.SIGINT, __signal_handler)

    # Second, get information about the job (node and token)
    node = None
    token = None
    
    cmd = [INTERPRETER,
           str(scripts_path + '/' + INFO_SCRIPT),
           job_id, app_path]
    return_code, stdout, stderr = _command_runner(cmd, login_info, modules=modules)
    if return_code != 0:
        __display_error(ERROR_INFO_JOB, return_code, stdout, stderr)

    # Parse the output
    out = stdout.splitlines()
    if NOT_RUNNING_KEYWORD in stdout:
        __display_warning(WARNING_NOTEBOOK_NOT_RUNNING)
        exit(0)
    elif SUCCESS_KEYWORD in stdout:
        for i in out[1:]:
            line = i.split(':')
            if line[0] == 'MASTER':
                node = line[1]
            elif line[0] == 'SERVER':
                try:
                    server_out = ' '.join(line[1:])
                    raw_token = re.search("token=\w*", server_out).group(0)
                    token = raw_token.split('=')[1]
                except AttributeError:
                    __display_error(ERROR_JUPYTER_SERVER)
    else:
        __display_error(ERROR_INFO_JOB, return_code, stdout, stderr)

    cmd = ['-L', f'{port_forward}:localhost:{port_forward}',
           'ssh', node,
           '-L', f'{port_forward}:localhost:8888']

    if utils.is_debug():
        print('****** DEBUG ******')
        print("\t -> Connecting to node: " + node)
        print("\t -> Token: " + token)
    
    _command_runner(cmd, login_info, blocking=False)

    time.sleep(5)  # Wait 5 seconds

    if web_browser is None:
        print(INFO_CONNECTION_ESTABLISHED)
        print(CONNECTION_URL.replace(':8888', f':{port_forward}') + token)
    else:
        print("Opening the " + web_browser + " browser with the connection URL.")
        if is_windows():
            cmd = ['cmd', '/c', 'start', web_browser]
        else:
            cmd = [web_browser]
        cmd = cmd + [CONNECTION_URL + token]
        return_code, stdout, stderr = _command_runner(cmd, login_info, remote=False)
        if return_code != 0:
            message = ERROR_BROWSER + '\n\n' \
                      + "Alternatively, please use the following URL to connect to the job.\n" \
                      + CONNECTION_URL + token
            __display_error(message, return_code, stdout, stderr)
        else:
            print(INFO_CONNECTION_ESTABLISHED)
            print(CONNECTION_URL + token)

    # Finally, wait for the CTRL+C signal
    print("Ready to work!")
    print("To force quit: CTRL + C")
    if is_windows():
        while True:
            # Waiting for signal
            try:
                time.sleep(5)
            except IOError:
                pass
    else:
        signal.pause()
    # The signal is captured and everything cleaned and canceled (if needed)


def cancel_job(scripts_path, job_ids, login_info, modules):
    """
    Cancel a list of notebook jobs running in the supercomputer.
    :param scripts_path: Path where the remote helper scripts are
    :param job_ids: List of job identifiers
    :return: None
    """
    cmd = [INTERPRETER,
           str(scripts_path + '/' + CANCEL_SCRIPT)] + job_ids
    return_code, stdout, stderr = _command_runner(cmd, login_info, modules=modules)
    if return_code != 0:
        __display_error(ERROR_CANCELLING_JOB, return_code, stdout, stderr)

    # Parse the output
    out = stdout.splitlines()
    if out[0] == SUCCESS_KEYWORD:
        print("Jobs successfully cancelled.")
    else:
        __display_error(ERROR_CANCELLING_JOB, return_code, stdout, stderr)


def _command_runner(cmd, login_info, modules=None, blocking=True, remote=True):
    """
    Run the command defined in the cmd list.
    Decodes the stdout and stderr following the DECODING_FORMAT.
    :param cmd: Command to execute as list.
    :param blocking: blocks until the subprocess finishes. Otherwise,
                     does not wait and appends the process to the global
                     alive processes list
    :param remote: Enable/Disable the execution in the remote supercomputer.
                   By default, prepend the needed SSH command.
                   (Uses the globals SSH, SESSION, USER_NAME and SUPERCOMPUTER).
    :return: return code, stdout, stderr | None if non blocking
    """
    global ALIVE_PROCESSES

    if remote:
        # Prepend the needed ssh command
        if is_windows():
            raise NotImplementedError()
        else:
            cmd = ' '.join(cmd)
            if modules:
                cmd = ';'.join([*modules, cmd])
            if 'ssh' in cmd:
                cmd = f"ssh {login_info} {cmd}"
            else:
                cmd = f"ssh {login_info} '{cmd}'"
    else:
        # Execute the command as requested
        pass

    if utils.is_debug():
        print('****** DEBUG ******')
        print("\t -> Command: " + cmd)
        print('********************')

    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if blocking:
        stdout, stderr = p.communicate()   # blocks until cmd is done
        stdout = stdout.decode(DECODING_FORMAT)
        stderr = stderr.decode(DECODING_FORMAT)
        return p.returncode, stdout, stderr
    else:
        ALIVE_PROCESSES.append(p)



#!/usr/bin/env python3

from pycompss_cli.core.arguments import parse_sys_argv
from pycompss_cli.core.actions_dispatcher import ActionsDispatcher

# Globals
DEBUG = False
LINE_LENGTH = 79
LINE = "-" * LINE_LENGTH


def main():
    """
    MAIN ENTRY POINT
    """
    global DEBUG
    arguments = parse_sys_argv()
    DEBUG = arguments.debug

    if DEBUG:
        print(LINE)
        print("Calling pycompss-cli for action: " + arguments.action)
        print("Arguments: " + str(arguments))

    ActionsDispatcher().run_action(arguments, DEBUG)

    if DEBUG:
        print(LINE)


if __name__ == "__main__":
    main()
